# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Node for an OPENQASM program.
"""
from ._node import Node


class Program(Node):
    """Node for an OPENQASM program.

    children is a list of nodes (statements).
    """

    def __init__(self, children):
        """Create the program node."""
        Node.__init__(self, 'program', children, None)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        string = ""
        for children in self.children:
            string += children.qasm(prec) + "\n"
        return string



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Base backend object for the unroller that raises BackendError.
"""
from ._backenderror import BackendError


class UnrollerBackend(object):
    """Backend for the unroller that raises BackendError.

    This backend also serves as a base class for other unroller backends.
    """

    def __init__(self, basis=None):
        """Setup this backend.

        basis is a list of operation name strings.
        """
        if basis:
            basis = []
        raise BackendError("Backend __init__ unimplemented")

    def set_basis(self, basis):
        """Declare the set of user-defined gates to emit.

        basis is a list of operation name strings.
        """
        raise BackendError("Backend set_basis unimplemented")

    def version(self, version):
        """Print the version string.

        v is a version number.
        """
        raise BackendError("Backend version unimplemented")

    def new_qreg(self, name, size):
        """Create a new quantum register.

        name = name of the register
        sz = size of the register
        """
        raise BackendError("Backend new_qreg unimplemented")

    def new_creg(self, name, size):
        """Create a new classical register.

        name = name of the register
        sz = size of the register
        """
        raise BackendError("Backend new_creg unimplemented")

    def define_gate(self, name, gatedata):
        """Define a new quantum gate.

        name is a string.
        gatedata is the AST node for the gate.
        """
        raise BackendError("Backend define_gate unimplemented")

    def u(self, arg, qubit, nested_scope=None):
        """Fundamental single qubit gate.

        arg is 3-tuple of Node expression objects.
        qubit is (regname,idx) tuple.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        raise BackendError("Backend u unimplemented")

    def cx(self, qubit0, qubit1):
        """Fundamental two qubit gate.

        qubit0 is (regname,idx) tuple for the control qubit.
        qubit1 is (regname,idx) tuple for the target qubit.
        """
        raise BackendError("Backend cx unimplemented")

    def measure(self, qubit, bit):
        """Measurement operation.

        qubit is (regname, idx) tuple for the input qubit.
        bit is (regname, idx) tuple for the output bit.
        """
        raise BackendError("Backend measure unimplemented")

    def barrier(self, qubitlists):
        """Barrier instruction.

        qubitlists is a list of lists of (regname, idx) tuples.
        """
        raise BackendError("Backend barrier unimplemented")

    def reset(self, qubit):
        """Reset instruction.

        qubit is a (regname, idx) tuple.
        """
        raise BackendError("Backend reset unimplemented")

    def set_condition(self, creg, cval):
        """Attach a current condition.

        creg is a name string.
        cval is the integer value for the test.
        """
        raise BackendError("Backend set_condition unimplemented")

    def drop_condition(self):
        """Drop the current condition."""
        raise BackendError("Backend drop_condition unimplemented")

    def start_gate(self, name, args, qubits, nested_scope=None):
        """Begin a custom gate.

        name is name string.
        args is list of Node expression objects.
        qubits is list of (regname, idx) tuples.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        raise BackendError("Backend start_gate unimplemented")

    def end_gate(self, name, args, qubits, nested_scope=None):
        """End a custom gate.

        name is name string.
        args is list of Node expression objects.
        qubits is list of (regname, idx) tuples.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        raise BackendError("Backend end_gate unimplemented")

    def get_output(self):
        """Returns the output generated by the backend.
        Depending on the type of Backend, the output could have different types.
        It must be called once the Qasm parsing has finished
        """
        raise BackendError("Backend get_output unimplemented")



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Node for an OPENQASM binary operation expression.
"""
from ._node import Node
from ._nodeexception import NodeException


class BinaryOp(Node):
    """Node for an OPENQASM binary operation expression.

    children[0] is the operation, as a character.
    children[1] is the left expression.
    children[2] is the right expression.
    """

    def __init__(self, children):
        """Create the binaryop node."""
        Node.__init__(self, 'binop', children, None)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        return "(" + self.children[1].qasm(prec) + self.children[0] + \
               self.children[2].qasm(prec) + ")"

    def latex(self, prec=15, nested_scope=None):
        """Return the corresponding math mode latex string."""
        return "(" + self.children[1].latex(prec, nested_scope) + \
            self.children[0] + \
            self.children[2].latex(prec, nested_scope) + ")"

    def real(self, nested_scope=None):
        """Return the correspond floating point number."""
        operation = self.children[0]
        lhs = self.children[1].real(nested_scope)
        rhs = self.children[2].real(nested_scope)
        if operation == '+':
            return lhs + rhs
        elif operation == '-':
            return lhs - rhs
        elif operation == '*':
            return lhs * rhs
        elif operation == '/':
            return lhs / rhs
        elif operation == '^':
            return lhs ** rhs
        else:
            raise NodeException("internal error: undefined binary op")



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Base node object for the OPENQASM syntax tree.
"""


class Node(object):
    """Base node object for the OPENQASM syntax tree."""

    def __init__(self, type, children=None, root=None):
        """Construct a new node object."""
        self.type = type
        if children:
            self.children = children
        else:
            self.children = []
        self.root = root
        # True if this node is an expression node, False otherwise
        self.expression = False

    def is_expression(self):
        """Return True if this is an expression node."""
        return self.expression

    def add_child(self, node):
        """Add a child node."""
        self.children.append(node)

    def to_string(self, indent):
        """Print with indent."""
        ind = indent * ' '
        if self.root:
            print(ind, self.type, '---', self.root)
        else:
            print(ind, self.type)
        indent = indent + 3
        ind = indent * ' '
        for children in self.children:
            if children is None:
                print("OOPS! type of parent is", type(self))
                print(self.children)
            if isinstance(children, str):
                print(ind, children)
            elif isinstance(children, int):
                print(ind, str(children))
            elif isinstance(children, float):
                print(ind, str(children))
            else:
                children.to_string(indent)



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Node for an OPENQASM creg statement.
"""
from ._node import Node


class Creg(Node):
    """Node for an OPENQASM creg statement.

    children[0] is an indexedid node.
    """

    def __init__(self, children):
        """Create the creg node."""
        Node.__init__(self, 'creg', children, None)
        # This is the indexed id, the full "id[n]" object
        self.id = children[0]
        # Name of the creg
        self.name = self.id.name
        # Source line number
        self.line = self.id.line
        # Source file name
        self.file = self.id.file
        # Size of the register
        self.index = self.id.index

    def to_string(self, indent):
        """Print the node data, with indent."""
        ind = indent * ' '
        print(ind, 'creg')
        self.children[0].to_string(indent + 3)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        return "creg " + self.id.qasm(prec) + ";"



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Object to represent a quantum circuit as a directed acyclic graph.

The nodes in the graph are either input/output nodes or operation nodes.
The operation nodes are elements of a basis that is part of the circuit.
The QASM definitions of the basis elements are carried with the circuit.
The edges correspond to qubits or bits in the circuit. A directed edge
from node A to node B means that the (qu)bit passes from the output of A
to the input of B. The object's methods allow circuits to be constructed,
composed, and modified. Some natural properties like depth can be computed
directly from the graph.
"""
import networkx as nx
import itertools
import copy
from ._dagcircuiterror import DAGCircuitError


class DAGCircuit:
    """
    Quantum circuit as a directed acyclic graph.

    There are 3 types of nodes in the graph: inputs, outputs, and operations.
    The nodes are connected by directed edges that correspond to qubits and
    bits.
    """

    def __init__(self):
        """Create an empty circuit."""
        # Map from a wire's name (reg,idx) to a Bool that is True if the
        # wire is a classical bit and False if the wire is a qubit.
        self.wire_type = {}

        # Map from wire names (reg,idx) to input nodes of the graph
        self.input_map = {}

        # Map from wire names (reg,idx) to output nodes of the graph
        self.output_map = {}

        # Running count of the total number of nodes
        self.node_counter = 0

        # Map of named operations in this circuit and their signatures.
        # The signature is an integer tuple (nq,nc,np) specifying the
        # number of input qubits, input bits, and real parameters.
        # The definition is external to the circuit object.
        self.basis = {}

        # Directed multigraph whose nodes are inputs, outputs, or operations.
        # Operation nodes have equal in- and out-degrees and carry
        # additional data about the operation, including the argument order
        # and parameter values.
        # Input nodes have out-degree 1 and output nodes have in-degree 1.
        # Edges carry wire labels (reg,idx) and each operation has
        # corresponding in- and out-edges with the same wire labels.
        self.multi_graph = nx.MultiDiGraph()

        # Map of qregs to sizes
        self.qregs = {}

        # Map of cregs to sizes
        self.cregs = {}

        # Map of user defined gates to ast nodes defining them
        self.gates = {}

        # Output precision for printing floats
        self.prec = 10

    def get_qubits(self):
        """Return a list of qubits as (qreg, index) pairs."""
        return [(k, i) for k, v in self.qregs.items() for i in range(v)]

    def rename_register(self, regname, newname):
        """Rename a classical or quantum register throughout the circuit.

        regname = existing register name string
        newname = replacement register name string
        """
        if regname == newname:
            return
        if newname in self.qregs or newname in self.cregs:
            raise DAGCircuitError("duplicate register name %s" % newname)
        if regname not in self.qregs and regname not in self.cregs:
            raise DAGCircuitError("no register named %s" % regname)
        iscreg = False
        if regname in self.qregs:
            self.qregs[newname] = self.qregs[regname]
            self.qregs.pop(regname, None)
            reg_size = self.qregs[newname]
        if regname in self.cregs:
            self.cregs[newname] = self.cregs[regname]
            self.cregs.pop(regname, None)
            reg_size = self.cregs[newname]
            iscreg = True
        for i in range(reg_size):
            self.wire_type[(newname, i)] = iscreg
            self.wire_type.pop((regname, i), None)
            self.input_map[(newname, i)] = self.input_map[(regname, i)]
            self.input_map.pop((regname, i), None)
            self.output_map[(newname, i)] = self.output_map[(regname, i)]
            self.output_map.pop((regname, i), None)
        # n node d = data
        for n, d in self.multi_graph.nodes_iter(data=True):
            if d["type"] == "in" or d["type"] == "out":
                if d["name"][0] == regname:
                    d["name"] = (newname, d["name"][1])
            elif d["type"] == "op":
                qa = []
                for a in d["qargs"]:
                    if a[0] == regname:
                        a = (newname, a[1])
                    qa.append(a)
                d["qargs"] = qa
                ca = []
                for a in d["cargs"]:
                    if a[0] == regname:
                        a = (newname, a[1])
                    ca.append(a)
                d["cargs"] = ca
                if d["condition"] is not None:
                    if d["condition"][0] == regname:
                        d["condition"] = (newname, d["condition"][1])
        # eX = edge, d= data
        for e1, e2, d in self.multi_graph.edges_iter(data=True):
            if d["name"][0] == regname:
                d["name"] = (newname, d["name"][1])

    def remove_all_ops_named(self, opname):
        """Remove all operation nodes with the given name."""
        nlist = self.get_named_nodes(opname)
        for n in nlist:
            self._remove_op_node(n)

    def deepcopy(self):
        """Return a deep copy of self."""
        return copy.deepcopy(self)

    def fs(self, number):
        """Format a float f as a string with self.prec digits."""
        fmt = "{0:0.%snumber}" % self.prec
        return fmt.format(number)

    def add_qreg(self, name, size):
        """Add all wires in a quantum register named name with size."""
        if name in self.qregs or name in self.cregs:
            raise DAGCircuitError("duplicate register name %s" % name)
        self.qregs[name] = size
        for j in range(size):
            self._add_wire((name, j))

    def add_creg(self, name, size):
        """Add all wires in a classical register named name with size."""
        if name in self.qregs or name in self.cregs:
            raise DAGCircuitError("duplicate register name %s" % name)
        self.cregs[name] = size
        for j in range(size):
            self._add_wire((name, j), True)

    def _add_wire(self, name, isClassical=False):
        """Add a qubit or bit to the circuit.

        name is a (string,int) tuple containing register name and index
        This adds a pair of in and out nodes connected by an edge.
        """
        if name not in self.wire_type:
            self.wire_type[name] = isClassical
            self.node_counter += 1
            self.input_map[name] = self.node_counter
            self.node_counter += 1
            self.output_map[name] = self.node_counter
            in_node = self.input_map[name]
            out_node = self.output_map[name]
            self.multi_graph.add_edge(in_node, out_node)
            self.multi_graph.node[in_node]["type"] = "in"
            self.multi_graph.node[out_node]["type"] = "out"
            self.multi_graph.node[in_node]["name"] = name
            self.multi_graph.node[out_node]["name"] = name
            self.multi_graph.edge[in_node][out_node][0]["name"] = name
        else:
            raise DAGCircuitError("duplicate wire %s" % name)

    def add_basis_element(self, name, number_qubits,
                          number_classical=0, number_parameters=0):
        """Add an operation to the basis.

        name is string label for operation
        number_qubits is number of qubit arguments
        number_classical is number of bit arguments
        number_parameters is number of real parameters

        The parameters (nq,nc,np) are ignored for the special case
        when name = "barrier". The barrier instruction has a variable
        number of qubit arguments.
        """
        if name not in self.basis:
            self.basis[name] = (
                number_qubits,
                number_classical,
                number_parameters)
        if name in self.gates:
            if self.gates[name]["n_args"] != number_parameters or \
                    self.gates[name]["n_bits"] != number_qubits or number_classical != 0:
                raise DAGCircuitError("gate data does not match "
                                      + "basis element specification")

    def add_gate_data(self, name, gatedata):
        """Add the definition of a gate.

        gatedata is dict with fields:
        "opaque" = True or False
        "n_args" = number of real parameters
        "n_bits" = number of qubits
        "args"   = list of parameter names
        "bits"   = list of qubit names
        "body"   = GateBody AST node
        """
        if name not in self.gates:
            self.gates[name] = gatedata
            if name in self.basis:
                if self.basis[name][0] != self.gates[name]["n_bits"] or \
                        self.basis[name][1] != 0 or \
                        self.basis[name][2] != self.gates[name]["n_args"]:
                    raise DAGCircuitError("gate data does not match "
                                          + "basis element specification")

    def _check_basis_data(self, name, qargs, cargs, params):
        """Check the arguments against the data for this operation.

        name is a string
        qargs is a list of tuples like ("q",0)
        cargs is a list of tuples like ("c",0)
        params is a list of strings that represent floats
        """
        # Check that we have this operation
        if name not in self.basis:
            raise DAGCircuitError("%s is not in the list of basis operations"
                                  % name)

        # Check the number of arguments matches the signature
        if name != "barrier":
            if len(qargs) != self.basis[name][0]:
                raise DAGCircuitError("incorrect number of qubits for %s"
                                      % name)
            if len(cargs) != self.basis[name][1]:
                raise DAGCircuitError("incorrect number of bits for %s"
                                      % name)
            if len(params) != self.basis[name][2]:
                raise DAGCircuitError("incorrect number of parameters for %s"
                                      % name)
        else:
            # "barrier" is a special case
            if len(qargs) == 0:
                raise DAGCircuitError("incorrect number of qubits for %s"
                                      % name)
            if len(cargs) != 0:
                raise DAGCircuitError("incorrect number of bits for %s"
                                      % name)
            if len(params) != 0:
                raise DAGCircuitError("incorrect number of parameters for %s"
                                      % name)

    def _check_condition(self, name, condition):
        """Verify that the condition is valid.

        name is a string used for error reporting
        condition is either None or a tuple (string,int) giving (creg,value)
        """
        # Verify creg exists
        if condition is not None and condition[0] not in self.cregs:
            raise DAGCircuitError("invalid creg in condition for %s" % name)

    def _check_bits(self, args, amap, bval):
        """Check the values of a list of (qu)bit arguments.

        For each element A of args, check that amap contains A and
        self.wire_type[A] equals bval.
        args is a list of (regname,idx) tuples
        amap is a dictionary keyed on (regname,idx) tuples
        bval is boolean
        """
        # Check for each wire
        for q in args:
            if q not in amap:
                raise DAGCircuitError("(qu)bit %s not found" % q)
            if self.wire_type[q] != bval:
                raise DAGCircuitError("expected wire type %s for %s"
                                      % (bval, q))

    def _bits_in_condition(self, cond):
        """Return a list of bits (regname,idx) in the given condition.

        cond is either None or a (regname,int) tuple specifying
        a classical if condition.
        """
        all_bits = []
        if cond is not None:
            all_bits.extend([(cond[0], j) for j in range(self.cregs[cond[0]])])
        return all_bits

    def _add_op_node(self, nname, nqargs, ncargs, nparams, ncondition):
        """Add a new operation node to the graph and assign properties.

        nname node name
        nqargs quantum arguments
        ncargs classical arguments
        nparams parameters
        ncondition classical condition (or None)
        """
        # Add a new operation node to the graph
        self.node_counter += 1
        self.multi_graph.add_node(self.node_counter)
        # Update that operation node's data
        self.multi_graph.node[self.node_counter]["type"] = "op"
        self.multi_graph.node[self.node_counter]["name"] = nname
        self.multi_graph.node[self.node_counter]["qargs"] = nqargs
        self.multi_graph.node[self.node_counter]["cargs"] = ncargs
        self.multi_graph.node[self.node_counter]["params"] = nparams
        self.multi_graph.node[self.node_counter]["condition"] = ncondition

    def apply_operation_back(self, name, qargs, cargs=[], params=[],
                             condition=None):
        """Apply an operation to the output of the circuit.

        name is a string
        qargs is a list of tuples like ("q",0)
        cargs is a list of tuples like ("c",0)
        params is a list of strings that represent floats
        condition is either None or a tuple (string,int) giving (creg,value)
        """
        all_cbits = self._bits_in_condition(condition)
        all_cbits.extend(cargs)

        self._check_basis_data(name, qargs, cargs, params)
        self._check_condition(name, condition)
        self._check_bits(qargs, self.output_map, False)
        self._check_bits(all_cbits, self.output_map, True)

        self._add_op_node(name, qargs, cargs, list(map(str, params)),
                          condition)
        # Add new in-edges from predecessors of the output nodes to the
        # operation node while deleting the old in-edges of the output nodes
        # and adding new edges from the operation node to each output node
        al = [qargs, all_cbits]
        for q in itertools.chain(*al):
            ie = self.multi_graph.predecessors(self.output_map[q])
            assert len(ie) == 1, "output node has multiple in-edges"
            self.multi_graph.add_edge(ie[0], self.node_counter, name=q)
            self.multi_graph.remove_edge(ie[0], self.output_map[q])
            self.multi_graph.add_edge(
                self.node_counter, self.output_map[q], name=q)

    def apply_operation_front(self, name, qargs, cargs=[], params=[],
                              condition=None):
        """Apply an operation to the input of the circuit.

        name is a string
        qargs is a list of strings like "q[0]"
        cargs is a list of strings like "c[0]"
        params is a list of strings that represent floats
        condition is either None or a tuple (string,int) giving (creg,value)
        """
        all_cbits = self._bits_in_condition(condition)
        all_cbits.extend(cargs)

        self._check_basis_data(name, qargs, cargs, params)
        self._check_condition(name, condition)
        self._check_bits(qargs, self.input_map, False)
        self._check_bits(all_cbits, self.input_map, True)

        self._add_op_node(name, qargs, cargs, list(map(str, params)),
                          condition)
        # Add new out-edges to successors of the input nodes from the
        # operation node while deleting the old out-edges of the input nodes
        # and adding new edges to the operation node from each input node
        al = [qargs, all_cbits]
        for q in itertools.chain(*al):
            ie = self.multi_graph.successors(self.input_map[q])
            assert len(ie) == 1, "input node has multiple out-edges"
            self.multi_graph.add_edge(self.node_counter, ie[0], name=q)
            self.multi_graph.remove_edge(self.input_map[q], ie[0])
            self.multi_graph.add_edge(
                self.input_map[q], self.node_counter, name=q)

    def _make_union_basis(self, input_circuit):
        """Return a new basis map.

        The new basis is a copy of self.basis with
        new elements of input_circuit.basis added.
        input_circuit is a DAGCircuit
        """
        union_basis = copy.deepcopy(self.basis)
        for g in input_circuit.basis:
            if g not in union_basis:
                union_basis[g] = input_circuit.basis[g]
            if union_basis[g] != input_circuit.basis[g]:
                raise DAGCircuitError("incompatible basis")
        return union_basis

    def _make_union_gates(self, input_circuit):
        """Return a new gates map.

        The new gates are a copy of self.gates with
        new elements of input_circuit.gates added.
        input_circuit is a DAGCircuit

        NOTE: gates in input_circuit that are also in self must
        be *identical* to the gates in self
        """
        union_gates = copy.deepcopy(self.gates)
        for k, v in input_circuit.gates.items():
            if k not in union_gates:
                union_gates[k] = v
            if union_gates[k]["opaque"] != input_circuit.gates[k]["opaque"] or\
               union_gates[k]["n_args"] != input_circuit.gates[k]["n_args"] or\
               union_gates[k]["n_bits"] != input_circuit.gates[k]["n_bits"] or\
               union_gates[k]["args"] != input_circuit.gates[k]["args"] or\
               union_gates[k]["bits"] != input_circuit.gates[k]["bits"]:
                raise DAGCircuitError("inequivalent gate definitions for %s"
                                      % k)
            if not union_gates[k]["opaque"] and \
               union_gates[k]["body"].qasm() != \
               input_circuit.gates[k]["body"].qasm():
                raise DAGCircuitError("inequivalent gate definitions for %s"
                                      % k)
        return union_gates

    def _check_wiremap_registers(self, wire_map, keyregs, valregs,
                                 valreg=True):
        """Check that wiremap neither fragments nor leaves duplicate registers.

        1. There are no fragmented registers. A register in keyregs
        is fragmented if not all of its (qu)bits are renamed by wire_map.
        2. There are no duplicate registers. A register is duplicate if
        it appears in both self and keyregs but not in wire_map.

        wire_map is a map from (regname,idx) in keyregs to (regname,idx)
        in valregs
        keyregs is a map from register names to sizes
        valregs is a map from register names to sizes
        valreg is a Bool, if False the method ignores valregs and does not
        add regs for bits in the wire_map image that don't appear in valregs
        Return the set of regs to add to self
        """
        add_regs = set([])
        reg_frag_chk = {}
        for k, v in keyregs.items():
            reg_frag_chk[k] = {j: False for j in range(v)}
        for k in wire_map.keys():
            if k[0] in keyregs:
                reg_frag_chk[k[0]][k[1]] = True
        for k, v in reg_frag_chk.items():
            rname = ",".join(map(str, k))
            s = set(v.values())
            if len(s) == 2:
                raise DAGCircuitError("wire_map fragments reg %s" % rname)
            elif s == set([False]):
                if k in self.qregs or k in self.cregs:
                    raise DAGCircuitError("unmapped duplicate reg %s" % rname)
                else:
                    # Add registers that appear only in keyregs
                    add_regs.add((k, keyregs[k]))
            else:
                if valreg:
                    # If mapping to a register not in valregs, add it.
                    # (k,0) exists in wire_map because wire_map doesn't
                    # fragment k
                    if not wire_map[(k, 0)][0] in valregs:
                        size = max(map(lambda x: x[1],
                                       filter(lambda x: x[0]
                                              == wire_map[(k, 0)][0],
                                              wire_map.values())))
                        add_regs.add((wire_map[(k, 0)][0], size + 1))
        return add_regs

    def _check_wiremap_validity(self, wire_map, keymap, valmap, input_circuit):
        """Check that the wiremap is consistent.

        Check that the wiremap refers to valid wires and that
        those wires have consistent types.

        wire_map is a map from (regname,idx) in keymap to (regname,idx)
        in valmap
        keymap is a map whose keys are wire_map keys
        valmap is a map whose keys are wire_map values
        input_circuit is a DAGCircuit
        """
        for k, v in wire_map.items():
            kname = ",".join(map(str, k))
            vname = ",".join(map(str, v))
            if k not in keymap:
                raise DAGCircuitError("invalid wire mapping key %s" % kname)
            if v not in valmap:
                raise DAGCircuitError("invalid wire mapping value %s"
                                      % vname)
            if input_circuit.wire_type[k] != self.wire_type[v]:
                raise DAGCircuitError("inconsistent wire_map at (%s,%s)"
                                      % (kname, vname))

    def _map_condition(self, wire_map, condition):
        """Use the wire_map dict to change the condition tuple's creg name.

        wire_map is map from wires to wires
        condition is a tuple (reg,int)
        Returns the new condition tuple
        """
        if condition is None:
            n_condition = None
        else:
            # Map the register name, using fact that registers must not be
            # fragmented by the wire_map (this must have been checked
            # elsewhere)
            bit0 = (condition[0], 0)
            n_condition = (wire_map.get(bit0, bit0)[0], condition[1])
        return n_condition

    def compose_back(self, input_circuit, wire_map={}):
        """Apply the input circuit to the output of this circuit.

        The two bases must be "compatible" or an exception occurs.
        A subset of input qubits of the input circuit are mapped
        to a subset of output qubits of this circuit.
        wire_map[input_qubit_to_input_circuit] = output_qubit_of_self
        """
        union_basis = self._make_union_basis(input_circuit)
        union_gates = self._make_union_gates(input_circuit)

        # Check the wire map for duplicate values
        if len(set(wire_map.values())) != len(wire_map):
            raise DAGCircuitError("duplicates in wire_map")

        add_qregs = self._check_wiremap_registers(wire_map,
                                                  input_circuit.qregs,
                                                  self.qregs)
        for register in add_qregs:
            self.add_qreg(register[0], register[1])

        add_cregs = self._check_wiremap_registers(wire_map,
                                                  input_circuit.cregs,
                                                  self.cregs)
        for register in add_cregs:
            self.add_creg(register[0], register[1])

        self._check_wiremap_validity(wire_map, input_circuit.input_map,
                                     self.output_map, input_circuit)

        # Compose
        self.basis = union_basis
        self.gates = union_gates
        topological_sort = nx.topological_sort(input_circuit.multi_graph)
        for node in topological_sort:
            nd = input_circuit.multi_graph.node[n]
            if nd["type"] == "in":
                # if in wire_map, get new name, else use existing name
                m_name = wire_map.get(nd["name"], nd["name"])
                # the mapped wire should already exist
                assert m_name in self.output_map, \
                    "wire (%s,%d) not in self" % (m_name[0], m_name[1])
                assert nd["name"] in input_circuit.wire_type, \
                    "inconsistent wire_type for (%s,%d) in input_circuit" \
                    % (nd["name"][0], nd["name"][1])
            elif nd["type"] == "out":
                # ignore output nodes
                pass
            elif nd["type"] == "op":
                condition = self._map_condition(wire_map, nd["condition"])
                self._check_condition(nd["name"], condition)
                m_qargs = list(map(lambda x: wire_map.get(x, x), nd["qargs"]))
                m_cargs = list(map(lambda x: wire_map.get(x, x), nd["cargs"]))
                self.apply_operation_back(nd["name"], m_qargs, m_cargs,
                                          nd["params"], condition)
            else:
                assert False, "bad node type %s" % nd["type"]

    def compose_front(self, input_circuit, wire_map={}):
        """Apply the input circuit to the input of this circuit.

        The two bases must be "compatible" or an exception occurs.
        A subset of output qubits of the input circuit are mapped
        to a subset of input qubits of
        this circuit.
        """
        union_basis = self._make_union_basis(input_circuit)
        union_gates = self._make_union_gates(input_circuit)

        # Check the wire map
        if len(set(wire_map.values())) != len(wire_map):
            raise DAGCircuitError("duplicates in wire_map")

        add_qregs = self._check_wiremap_registers(wire_map,
                                                  input_circuit.qregs,
                                                  self.qregs)
        for r in add_qregs:
            self.add_qreg(r[0], r[1])

        add_cregs = self._check_wiremap_registers(wire_map,
                                                  input_circuit.cregs,
                                                  self.cregs)
        for r in add_cregs:
            self.add_creg(r[0], r[1])

        self._check_wiremap_validity(wire_map, input_circuit.output_map,
                                     self.input_map, input_circuit)

        # Compose
        self.basis = union_basis
        self.gates = union_gates
        ts = nx.topological_sort(input_circuit.multi_graph, reverse=True)
        for n in ts:
            nd = input_circuit.multi_graph.node[n]
            if nd["type"] == "out":
                # if in wire_map, get new name, else use existing name
                m_name = wire_map.get(nd["name"], nd["name"])
                # the mapped wire should already exist
                assert m_name in self.input_map, \
                    "wire (%s,%d) not in self" % (m_name[0], m_name[1])
                assert nd["name"] in input_circuit.wire_type, \
                    "inconsistent wire_type for (%s,%d) in input_circuit" \
                    % (nd["name"][0], nd["name"][1])
            elif nd["type"] == "in":
                # ignore input nodes
                pass
            elif nd["type"] == "op":
                condition = self._map_condition(wire_map, nd["condition"])
                self._check_condition(nd["name"], condition)
                m_qargs = list(map(lambda x: wire_map.get(x, x), nd["qargs"]))
                m_cargs = list(map(lambda x: wire_map.get(x, x), nd["cargs"]))
                self.apply_operation_front(nd["name"], m_qargs, m_cargs,
                                           nd["params"], condition)
            else:
                assert False, "bad node type %s" % nd["type"]

    def size(self):
        """Return the number of operations."""
        return self.multi_graph.order() - 2 * len(self.wire_type)

    def depth(self):
        """Return the circuit depth."""
        assert nx.is_directed_acyclic_graph(self.multi_graph), "not a DAG"
        return nx.dag_longest_path_length(self.multi_graph) - 1

    def width(self):
        """Return the total number of qubits used by the circuit."""
        return len(self.wire_type) - self.num_cbits()

    def num_cbits(self):
        """Return the total number of bits used by the circuit."""
        return list(self.wire_type.values()).count(True)

    def num_tensor_factors(self):
        """Compute how many components the circuit can decompose into."""
        return nx.number_weakly_connected_components(self.multi_graph)

    def _gate_string(self, name):
        """Return a QASM string for the named gate."""
        out = ""
        if self.gates[name]["opaque"]:
            out = "opaque " + name
        else:
            out = "gate " + name
        if self.gates[name]["n_args"] > 0:
            out += "(" + ",".join(self.gates[name]["args"]) + ")"
        out += " " + ",".join(self.gates[name]["bits"])
        if self.gates[name]["opaque"]:
            out += ";"
        else:
            out += "\n{\n" + self.gates[name]["body"].qasm() + "}"
        return out

    def qasm(self, decls_only=False, add_swap=False,
             no_decls=False, qeflag=False, aliases=None):
        """Return a string containing QASM for this circuit.

        if qeflag is True, add a line to include "qelib1.inc"
        and only generate gate code for gates not in qelib1.

        if no_decls is True, only print the instructions.

        if aliases is not None, aliases contains a dict mapping
        the current qubits in the circuit to new qubit names.
        We will deduce the register names and sizes from aliases.

        if decls_only is True, only print the declarations.

        if add_swap is True, add the definition of swap in terms of
        cx if necessary.
        """
        # Rename qregs if necessary
        if aliases:
            qregdata = {}
            for q in aliases.values():
                if q[0] not in qregdata:
                    qregdata[q[0]] = q[1] + 1
                elif qregdata[q[0]] < q[1] + 1:
                    qregdata[q[0]] = q[1] + 1
        else:
            qregdata = self.qregs
        # Write top matter
        if no_decls:
            out = ""
        else:
            printed_gates = []
            out = "OPENQASM 2.0;\n"
            if qeflag:
                out += "include \"qelib1.inc\";\n"
            for k, v in sorted(qregdata.items()):
                out += "qreg %s[%d];\n" % (k, v)
            for k, v in sorted(self.cregs.items()):
                out += "creg %s[%d];\n" % (k, v)
            omit = ["U", "CX", "measure", "reset", "barrier"]
            if qeflag:
                qelib = ["u3", "u2", "u1", "cx", "id", "x", "y", "z", "h",
                         "s", "sdg", "t", "tdg", "cz", "cy", "ccx", "cu1",
                         "cu3"]
                omit.extend(qelib)
                printed_gates.extend(qelib)
            for k in self.basis.keys():
                if k not in omit:
                    if not self.gates[k]["opaque"]:
                        calls = self.gates[k]["body"].calls()
                        for c in calls:
                            if c not in printed_gates:
                                out += self._gate_string(c) + "\n"
                                printed_gates.append(c)
                    if k not in printed_gates:
                        out += self._gate_string(k) + "\n"
                        printed_gates.append(k)
            if add_swap and not qeflag and "cx" not in self.basis:
                out += "gate cx a,b { CX a,b; }\n"
            if add_swap and "swap" not in self.basis:
                out += "gate swap a,b { cx a,b; cx b,a; cx a,b; }\n"
        # Write the instructions
        if not decls_only:
            ts = nx.topological_sort(self.multi_graph)
            for n in ts:
                nd = self.multi_graph.node[n]
                if nd["type"] == "op":
                    if nd["condition"] is not None:
                        out += "if(%s==%d) " \
                               % (nd["condition"][0], nd["condition"][1])
                    if len(nd["cargs"]) == 0:
                        nm = nd["name"]
                        if aliases:
                            qarglist = map(lambda x: aliases[x], nd["qargs"])
                        else:
                            qarglist = nd["qargs"]
                        qarg = ",".join(map(lambda x: "%s[%d]" % (x[0], x[1]),
                                            qarglist))
                        if len(nd["params"]) > 0:
                            param = ",".join(nd["params"])
                            out += "%s(%s) %s;\n" % (nm, param, qarg)
                        else:
                            out += "%s %s;\n" % (nm, qarg)
                    else:
                        if nd["name"] == "measure":
                            assert len(nd["cargs"]) == 1 and \
                                len(nd["qargs"]) == 1 and \
                                len(nd["params"]) == 0, "bad node data"
                            qname = nd["qargs"][0][0]
                            qindex = nd["qargs"][0][1]
                            if aliases:
                                newq = aliases[(qname, qindex)]
                                qname = newq[0]
                                qindex = newq[1]
                            out += "measure %s[%d] -> %s[%d];\n" \
                                   % (qname,
                                      qindex,
                                      nd["cargs"][0][0],
                                      nd["cargs"][0][1])
                        else:
                            assert False, "bad node data"
        return out

    def _check_wires_list(self, wires, name, input_circuit):
        """Check that a list of wires satisfies some conditions.

        The wires give an order for (qu)bits in the input circuit
        that is replacing the named operation.
        - no duplicate names
        - correct length for named operation
        - elements are wires of input_circuit
        Raises an exception otherwise.
        """
        if len(set(wires)) != len(wires):
            raise DAGCircuitError("duplicate wires")

        wire_tot = self.basis[name][0] + self.basis[name][1]
        if len(wires) != wire_tot:
            raise DAGCircuitError("expected %d wires, got %d"
                                  % (wire_tot, len(wires)))

        for w in wires:
            if w not in input_circuit.wire_type:
                raise DAGCircuitError("wire (%s,%d) not in input circuit"
                                      % (w[0], w[1]))

    def _make_pred_succ_maps(self, n):
        """Return predecessor and successor dictionaries.

        These map from wire names to predecessor and successor
        nodes for the operation node n in self.multi_graph.
        """
        pred_map = {e[2]['name']: e[0] for e in
                    self.multi_graph.in_edges_iter(nbunch=n, data=True)}
        succ_map = {e[2]['name']: e[1] for e in
                    self.multi_graph.out_edges_iter(nbunch=n, data=True)}
        return pred_map, succ_map

    def _full_pred_succ_maps(self, pred_map, succ_map, input_circuit,
                             wire_map):
        """Map all wires of the input circuit.

        Map all wires of the input circuit to predecessor and
        successor nodes in self, keyed on wires in self.

        pred_map, succ_map dicts come from _make_pred_succ_maps
        input_circuit is the input circuit
        wire_map is the wire map from wires of input_circuit to wires of self
        returns full_pred_map, full_succ_map
        """
        full_pred_map = {}
        full_succ_map = {}
        for w in input_circuit.input_map:
            # If w is wire mapped, find the corresponding predecessor
            # of the node
            if w in wire_map:
                full_pred_map[wire_map[w]] = pred_map[wire_map[w]]
                full_succ_map[wire_map[w]] = succ_map[wire_map[w]]
            else:
                # Otherwise, use the corresponding output nodes of self
                # and compute the predecessor.
                full_succ_map[w] = self.output_map[w]
                full_pred_map[w] = self.multi_graph.predecessors(self.output_map[w])[
                    0]
                assert len(self.multi_graph.predecessors(self.output_map[w])) == 1,\
                    "too many predecessors for (%s,%d) output node" % (
                        w[0], w[1])
        return full_pred_map, full_succ_map

    def substitute_circuit_all(self, name, input_circuit, wires=[]):
        """Replace every occurrence of named operation with input_circuit."""
        if name not in self.basis:
            raise DAGCircuitError("%s is not in the list of basis operations"
                                  % name)

        self._check_wires_list(wires, name, input_circuit)
        union_basis = self._make_union_basis(input_circuit)
        union_gates = self._make_union_gates(input_circuit)

        # Create a proxy wire_map to identify fragments and duplicates
        # and determine what registers need to be added to self
        proxy_map = {w: ("", 0) for w in wires}
        add_qregs = self._check_wiremap_registers(proxy_map,
                                                  input_circuit.qregs,
                                                  {}, False)
        for r in add_qregs:
            self.add_qreg(r[0], r[1])

        add_cregs = self._check_wiremap_registers(proxy_map,
                                                  input_circuit.cregs,
                                                  {}, False)
        for r in add_cregs:
            self.add_creg(r[0], r[1])

        # Iterate through the nodes of self and replace the selected nodes
        # by iterating through the input_circuit, constructing and
        # checking the validity of the wire_map for each replacement
        # NOTE: We do not replace conditioned gates. One way to implement
        #       this later is to add or update the conditions of each gate
        #       that we add from the input_circuit.
        self.basis = union_basis
        self.gates = union_gates
        ts = nx.topological_sort(self.multi_graph)
        for n in ts:
            nd = self.multi_graph.node[n]
            if nd["type"] == "op" and nd["name"] == name:
                if nd["condition"] is None:
                    wire_map = {k: v for k, v in zip(wires,
                                                     [i for s in [nd["qargs"], nd["cargs"]]
                                                      for i in s])}
                    self._check_wiremap_validity(wire_map, wires,
                                                 self.input_map, input_circuit)
                    pred_map, succ_map = self._make_pred_succ_maps(n)
                    full_pred_map, full_succ_map = \
                        self._full_pred_succ_maps(pred_map, succ_map,
                                                  input_circuit, wire_map)
                    # Now that we know the connections, delete node
                    self.multi_graph.remove_node(n)
                    # Iterate over nodes of input_circuit
                    tsin = nx.topological_sort(input_circuit.multi_graph)
                    for m in tsin:
                        md = input_circuit.multi_graph.node[m]
                        if md["type"] == "op":
                            # Insert a new node
                            condition = self._map_condition(wire_map,
                                                            md["condition"])
                            m_qargs = list(map(lambda x: wire_map.get(x, x),
                                               md["qargs"]))
                            m_cargs = list(map(lambda x: wire_map.get(x, x),
                                               md["cargs"]))
                            self._add_op_node(md["name"], m_qargs, m_cargs,
                                              md["params"], condition)
                            # Add edges from predecessor nodes to new node
                            # and update predecessor nodes that change
                            all_cbits = self._bits_in_condition(condition)
                            all_cbits.extend(m_cargs)
                            al = [m_qargs, all_cbits]
                            for q in itertools.chain(*al):
                                self.multi_graph.add_edge(full_pred_map[q],
                                                          self.node_counter, name=q)
                                full_pred_map[q] = copy.copy(self.node_counter)
                    # Connect all predecessors and successors, and remove
                    # residual edges between input and output nodes
                    for w in full_pred_map.keys():
                        self.multi_graph.add_edge(full_pred_map[w], full_succ_map[w],
                                                  name=w)
                        o_pred = self.multi_graph.predecessors(
                            self.output_map[w])
                        if len(o_pred) > 1:
                            assert len(o_pred) == 2, \
                                "expected 2 predecessors here"
                            p = list(filter(lambda x: x != full_pred_map[w],
                                            o_pred))
                            assert len(p) == 1, \
                                "expected 1 predecessor to pass filter"
                            self.multi_graph.remove_edge(
                                p[0], self.output_map[w])

    def substitute_circuit_one(self, node, input_circuit, wires=[]):
        """Replace one node with input_circuit.

        node is a reference to a node of self.multi_graph of type "op"
        input_circuit is a DAGCircuit
        """
        nd = self.multi_graph.node[node]

        # TODO: reuse common code in substitute_circuit_one and _all

        name = nd["name"]
        self._check_wires_list(wires, name, input_circuit)
        union_basis = self._make_union_basis(input_circuit)
        union_gates = self._make_union_gates(input_circuit)

        # Create a proxy wire_map to identify fragments and duplicates
        # and determine what registers need to be added to self
        proxy_map = {w: ("", 0) for w in wires}
        add_qregs = self._check_wiremap_registers(proxy_map,
                                                  input_circuit.qregs,
                                                  {}, False)
        for r in add_qregs:
            self.add_qreg(r[0], r[1])

        add_cregs = self._check_wiremap_registers(proxy_map,
                                                  input_circuit.cregs,
                                                  {}, False)
        for r in add_cregs:
            self.add_creg(r[0], r[1])

        # Replace the node by iterating through the input_circuit.
        # Constructing and checking the validity of the wire_map.
        # NOTE: We do not replace conditioned gates. One way to implement
        #       later is to add or update the conditions of each gate we add
        #       from the input_circuit.
        self.basis = union_basis
        self.gates = union_gates

        if nd["type"] != "op":
            raise DAGCircuitError("expected node type \"op\", got %s"
                                  % nd["type"])

        if nd["condition"] is None:
            wire_map = {k: v for k, v in zip(wires,
                                             [i for s in [nd["qargs"],
                                                          nd["cargs"]]
                                              for i in s])}
            self._check_wiremap_validity(wire_map, wires,
                                         self.input_map, input_circuit)
            pred_map, succ_map = self._make_pred_succ_maps(node)
            full_pred_map, full_succ_map = \
                self._full_pred_succ_maps(pred_map, succ_map,
                                          input_circuit, wire_map)
            # Now that we know the connections, delete node
            self.multi_graph.remove_node(node)
            # Iterate over nodes of input_circuit
            tsin = nx.topological_sort(input_circuit.multi_graph)
            for m in tsin:
                md = input_circuit.multi_graph.node[m]
                if md["type"] == "op":
                    # Insert a new node
                    condition = self._map_condition(wire_map, md["condition"])
                    m_qargs = list(map(lambda x: wire_map.get(x, x),
                                       md["qargs"]))
                    m_cargs = list(map(lambda x: wire_map.get(x, x),
                                       md["cargs"]))
                    self._add_op_node(md["name"], m_qargs, m_cargs,
                                      md["params"], condition)
                    # Add edges from predecessor nodes to new node
                    # and update predecessor nodes that change
                    all_cbits = self._bits_in_condition(condition)
                    all_cbits.extend(m_cargs)
                    al = [m_qargs, all_cbits]
                    for q in itertools.chain(*al):
                        self.multi_graph.add_edge(full_pred_map[q], self.node_counter,
                                                  name=q)
                        full_pred_map[q] = copy.copy(self.node_counter)
            # Connect all predecessors and successors, and remove
            # residual edges between input and output nodes
            for w in full_pred_map.keys():
                self.multi_graph.add_edge(
                    full_pred_map[w], full_succ_map[w], name=w)
                o_pred = self.multi_graph.predecessors(self.output_map[w])
                if len(o_pred) > 1:
                    assert len(o_pred) == 2, "expected 2 predecessors here"
                    p = list(filter(lambda x: x != full_pred_map[w], o_pred))
                    assert len(p) == 1, "expected 1 predecessor to pass filter"
                    self.multi_graph.remove_edge(p[0], self.output_map[w])

    def get_named_nodes(self, name):
        """Return a list of "op" nodes with the given name."""
        nlist = []
        if name not in self.basis:
            raise DAGCircuitError("%s is not in the list of basis operations"
                                  % name)

        # Iterate through the nodes of self in topological order
        ts = nx.topological_sort(self.multi_graph)
        for n in ts:
            nd = self.multi_graph.node[n]
            if nd["type"] == "op" and nd["name"] == name:
                nlist.append(n)
        return nlist

    def _remove_op_node(self, n):
        """Remove an operation node n.

        Add edges from predecessors to successors.
        """
        pred_map, succ_map = self._make_pred_succ_maps(n)
        self.multi_graph.remove_node(n)
        for w in pred_map.keys():
            self.multi_graph.add_edge(pred_map[w], succ_map[w], name=w)

    def remove_ancestors_of(self, node):
        """Remove all of the ancestor operation nodes of node."""
        anc = nx.ancestors(self.multi_graph, node)
        # TODO: probably better to do all at once using
        # multi_graph.remove_nodes_from; same for related functions ...
        for n in anc:
            nd = self.multi_graph.node[n]
            if nd["type"] == "op":
                self._remove_op_node(n)

    def remove_descendants_of(self, node):
        """Remove all of the descendant operation nodes of node."""
        dec = nx.descendants(self.multi_graph, node)
        for n in dec:
            nd = self.multi_graph.node[n]
            if nd["type"] == "op":
                self._remove_op_node(n)

    def remove_nonancestors_of(self, node):
        """Remove all of the non-ancestors operation nodes of node."""
        anc = nx.ancestors(self.multi_graph, node)
        comp = list(set(self.multi_graph.nodes()) - set(anc))
        for n in comp:
            nd = self.multi_graph.node[n]
            if nd["type"] == "op":
                self._remove_op_node(n)

    def remove_nondescendants_of(self, node):
        """Remove all of the non-descendants operation nodes of node."""
        dec = nx.descendants(self.multi_graph, node)
        comp = list(set(self.multi_graph.nodes()) - set(dec))
        for n in comp:
            nd = self.multi_graph.node[n]
            if nd["type"] == "op":
                self._remove_op_node(n)

    def layers(self):
        """Return a list of layers for all d layers of this circuit.

        A layer is a circuit whose gates act on disjoint qubits, i.e.
        a layer has depth 1. The total number of layers equals the
        circuit depth d. The layers are indexed from 0 to d-1 with the
        earliest layer at index 0. The layers are constructed using a
        greedy algorithm. Each returned layer is a dict containing
        {"graph": circuit graph, "partition": list of qubit lists}.


        TODO: Gates that use the same cbits will end up in different
        layers as this is currently implemented. This may not be
        the desired behavior.
        """
        layers_list = []
        # node_map contains an input node or previous layer node for
        # each wire in the circuit.
        node_map = copy.deepcopy(self.input_map)
        # wires_with_ops_remaining is a set of wire names that have
        # operations we still need to assign to layers
        wires_with_ops_remaining = set(self.input_map.keys())
        while wires_with_ops_remaining:
            # Create a new circuit graph and populate with regs and basis
            new_layer = DAGCircuit()
            for k, v in self.qregs.items():
                new_layer.add_qreg(k, v)
            for k, v in self.cregs.items():
                new_layer.add_creg(k, v)
            new_layer.basis = copy.deepcopy(self.basis)
            new_layer.gates = copy.deepcopy(self.gates)
            # Save the support of each operation we add to the layer
            support_list = []
            # Determine what operations to add in this layer
            # ops_touched is a map from operation nodes touched in this
            # iteration to the set of their unvisited input wires. When all
            # of the inputs of a touched node are visited, the node is a
            # foreground node we can add to the current layer.
            ops_touched = {}
            wires_loop = list(wires_with_ops_remaining)
            emit = False
            for w in wires_loop:
                oe = list(filter(lambda x: x[2]["name"] == w,
                                 self.multi_graph.out_edges(nbunch=[node_map[w]],
                                                            data=True)))
                assert len(oe) == 1, "should only be one out-edge per (qu)bit"
                nxt_nd_idx = oe[0][1]
                nxt_nd = self.multi_graph.node[nxt_nd_idx]
                # If we reach an output node, we are done with this wire.
                if nxt_nd["type"] == "out":
                    wires_with_ops_remaining.remove(w)
                # Otherwise, we are somewhere inside the circuit
                elif nxt_nd["type"] == "op":
                    # Operation data
                    qa = copy.copy(nxt_nd["qargs"])
                    ca = copy.copy(nxt_nd["cargs"])
                    pa = copy.copy(nxt_nd["params"])
                    co = copy.copy(nxt_nd["condition"])
                    cob = self._bits_in_condition(co)
                    # First time we see an operation, add to ops_touched
                    if nxt_nd_idx not in ops_touched:
                        ops_touched[nxt_nd_idx] = set(qa) | set(ca) | set(cob)
                    # Mark inputs visited by deleting from set
                    # NOTE: expect trouble with if(c==1) measure q -> c;
                    assert w in ops_touched[nxt_nd_idx], "expected wire"
                    ops_touched[nxt_nd_idx].remove(w)
                    # Node becomes "foreground" if set becomes empty,
                    # i.e. every input is available for this operation
                    if not ops_touched[nxt_nd_idx]:
                        # Add node to new_layer
                        new_layer.apply_operation_back(nxt_nd["name"],
                                                       qa, ca, pa, co)
                        # Update node_map to point to this op
                        for v in itertools.chain(qa, ca, cob):
                            node_map[v] = nxt_nd_idx
                        # Add operation to partition
                        if nxt_nd["name"] != "barrier":
                            # support_list.append(list(set(qa) | set(ca) |
                            #                          set(cob)))
                            support_list.append(list(set(qa)))
                        emit = True
            if emit:
                l_dict = {"graph": new_layer, "partition": support_list}
                layers_list.append(l_dict)
                emit = False
            else:
                assert not wires_with_ops_remaining, "not finished but empty?"
        return layers_list

    def serial_layers(self):
        """Return a list of layers for all gates of this circuit.

        A serial layer is a circuit with one gate. The layers have the
        same structure as in layers().
        """
        layers_list = []
        ts = nx.topological_sort(self.multi_graph)
        for n in ts:
            nxt_nd = self.multi_graph.node[n]
            if nxt_nd["type"] == "op":
                new_layer = DAGCircuit()
                for k, v in self.qregs.items():
                    new_layer.add_qreg(k, v)
                for k, v in self.cregs.items():
                    new_layer.add_creg(k, v)
                new_layer.basis = copy.deepcopy(self.basis)
                new_layer.gates = copy.deepcopy(self.gates)
                # Save the support of the operation we add to the layer
                support_list = []
                # Operation data
                qa = copy.copy(nxt_nd["qargs"])
                ca = copy.copy(nxt_nd["cargs"])
                pa = copy.copy(nxt_nd["params"])
                co = copy.copy(nxt_nd["condition"])
                cob = self._bits_in_condition(co)
                # Add node to new_layer
                new_layer.apply_operation_back(nxt_nd["name"],
                                               qa, ca, pa, co)
                # Add operation to partition
                if nxt_nd["name"] != "barrier":
                    # support_list.append(list(set(qa) | set(ca) | set(cob)))
                    support_list.append(list(set(qa)))
                l_dict = {"graph": new_layer, "partition": support_list}
                layers_list.append(l_dict)
        return layers_list

    def collect_runs(self, namelist):
        """Return a set of runs of "op" nodes with the given names.

        For example, "... h q[0]; cx q[0],q[1]; cx q[0],q[1]; h q[1]; .."
        would produce the tuple of cx nodes as an element of the set returned
        from a call to collect_runs(["cx"]). If instead the cx nodes were
        "cx q[0],q[1]; cx q[1],q[0];", the method would still return the
        pair in a tuple. The namelist can contain names that are not
        in the circuit's basis.

        Nodes must have only one successor to continue the run.
        """
        group_list = []

        # Iterate through the nodes of self in topological order
        # and form tuples containing sequences of gates
        # on the same qubit(s).
        ts = nx.topological_sort(self.multi_graph)
        nodes_seen = dict(zip(ts, [False] * len(ts)))
        for node in ts:
            nd = self.multi_graph.node[node]
            if nd["type"] == "op" and nd["name"] in namelist \
               and not nodes_seen[node]:
                group = [node]
                nodes_seen[node] = True
                s = self.multi_graph.successors(node)
                while len(s) == 1 and \
                        self.multi_graph.node[s[0]]["type"] == "op" and \
                        self.multi_graph.node[s[0]]["name"] in namelist:
                    group.append(s[0])
                    nodes_seen[s[0]] = True
                    s = self.multi_graph.successors(s[0])
                if len(group) > 1:
                    group_list.append(tuple(group))
        return set(group_list)

    def count_ops(self):
        """Count the occurrences of operation names.

        Returns a dictionary of counts keyed on the operation name.
        """
        op_dict = {}
        ts = nx.topological_sort(self.multi_graph)
        for node in ts:
            nd = self.multi_graph.node[node]
            name = nd["name"]
            if nd["type"] == "op":
                if name not in op_dict:
                    op_dict[name] = 1
                else:
                    op_dict[name] += 1
        return op_dict

    def property_summary(self):
        """Return a dictionary of circuit properties."""
        summary = {"size": self.size(),
                   "depth": self.depth(),
                   "width": self.width(),
                   "bits": self.num_cbits(),
                   "factors": self.num_tensor_factors(),
                   "operations": self.count_ops()}
        return summary



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Hadamard gate.
"""
from qiskit import QuantumCircuit
from qiskit import QuantumRegister
from qiskit import Gate
from qiskit import CompositeGate
from qiskit import InstructionSet
from qiskit.extensions.standard import header


class HGate(Gate):
    """Hadamard gate."""

    def __init__(self, qubit, circ=None):
        """Create new Hadamard gate."""
        super(HGate, self).__init__("h", [], [qubit], circ)

    def qasm(self):
        """Return OPENQASM string."""
        qubit = self.arg[0]
        return self._qasmif("h %s[%d];" % (qubit[0].name, qubit[1]))

    def inverse(self):
        """Invert this gate."""
        return self  # self-inverse

    def reapply(self, circ):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circ.h(self.arg[0]))


def h(self, q):
    """Apply H to q."""
    if isinstance(q, QuantumRegister):
        gs = InstructionSet()
        for j in range(q.size):
            gs.add(self.h((q, j)))
        return gs
    else:
        self._check_qubit(q)
        return self._attach(HGate(q, self))


QuantumCircuit.h = h
CompositeGate.h = h



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Classical register reference object.
"""
from ._register import Register


class ClassicalRegister(Register):
    """Implement a classical register."""

    def qasm(self):
        """Return OPENQASM string for this register."""
        return "creg %s[%d];" % (self.name, self.size)

    def __str__(self):
        """Return a string representing the register."""
        return "ClassicalRegister(%s,%d)" % (self.name, self.size)

    def __len__(self):
        """Return a int representing the size."""
        return self.size



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Pauli Y (bit-phase-flip) gate.
"""
from qiskit import QuantumRegister
from qiskit import QuantumCircuit
from qiskit import Gate
from qiskit import CompositeGate
from qiskit import InstructionSet
from qiskit.extensions.standard import header


class YGate(Gate):
    """Pauli Y (bit-phase-flip) gate."""

    def __init__(self, qubit, circ=None):
        """Create new Y gate."""
        super(YGate, self).__init__("y", [], [qubit], circ)

    def qasm(self):
        """Return OPENQASM string."""
        qubit = self.arg[0]
        return self._qasmif("y %s[%d];" % (qubit[0].name, qubit[1]))

    def inverse(self):
        """Invert this gate."""
        return self  # self-inverse

    def reapply(self, circuit):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circuit.y(self.arg[0]))


def y(self, quantum_register):
    """Apply Y to q."""
    if isinstance(quantum_register, QuantumRegister):
        intructions = InstructionSet()
        for register in range(quantum_register.size):
            intructions.add(self.y((quantum_register, register)))
        return intructions
    else:
        self._check_qubit(quantum_register)
        return self._attach(YGate(quantum_register, self))


QuantumCircuit.y = y
CompositeGate.y = y



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Rotation around the x-axis.
"""
from qiskit import QuantumRegister
from qiskit import QuantumCircuit
from qiskit import Gate
from qiskit import InstructionSet
from qiskit import CompositeGate
from qiskit.extensions.standard import header


class RXGate(Gate):
    """rotation around the x-axis."""

    def __init__(self, theta, qubit, circ=None):
        """Create new rx single qubit gate."""
        super(RXGate, self).__init__("rx", [theta], [qubit], circ)

    def qasm(self):
        """Return OPENQASM string."""
        qubit = self.arg[0]
        theta = self.param[0]
        return self._qasmif("rx(%.15f) %s[%d];" % (theta, qubit[0].name,
                                                   qubit[1]))

    def inverse(self):
        """Invert this gate.

        rx(theta)^dagger = rx(-theta)
        """
        self.param[0] = -self.param[0]
        return self

    def reapply(self, circ):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circ.rx(self.param[0], self.arg[0]))


def rx(self, theta, q):
    """Apply rx to q."""
    if isinstance(q, QuantumRegister):
        gs = InstructionSet()
        for j in range(q.sz):
            gs.add(self.rx(theta, (q, j)))
        return gs
    else:
        self._check_qubit(q)
        return self._attach(RXGate(theta, q, self))


QuantumCircuit.rx = rx
CompositeGate.rx = rx



from ._coupling import Coupling, coupling_dict2list, coupling_list2dict
from ._mapping import swap_mapper, direction_mapper, cx_cancellation, optimize_1q_gates
from ._compiling import two_qubit_kak, euler_angles_1q
from ._mappererror import MapperError



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
T=sqrt(S) phase gate or its inverse.
"""
import math
from qiskit import QuantumRegister
from qiskit import QuantumCircuit
from qiskit import CompositeGate
from qiskit import InstructionSet
from qiskit.extensions.standard import header
from qiskit.extensions.standard import u1


class TGate(CompositeGate):
    """T=sqrt(S) Clifford phase gate or its inverse."""

    def __init__(self, qubit, circ=None):
        """Create new T gate."""
        super(TGate, self).__init__("t", [], [qubit], circ)
        self.u1(math.pi / 4.0, qubit)

    def reapply(self, circ):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circ.t(self.arg[0]))

    def qasm(self):
        """Return OPENQASM string."""
        qubit = self.data[0].arg[0]
        phi = self.data[0].param[0]
        if phi > 0:
            return self.data[0]._qasmif("t %s[%d];" % (qubit[0].name, qubit[1]))
        else:
            return self.data[0]._qasmif("tdg %s[%d];" % (qubit[0].name, qubit[1]))


def t(self, q):
    """Apply T to q."""
    if isinstance(q, QuantumRegister):
        gs = InstructionSet()
        for j in range(q.size):
            gs.add(self.t((q, j)))
        return gs
    else:
        self._check_qubit(q)
        return self._attach(TGate(q, self))


def tdg(self, q):
    """Apply Tdg to q."""
    return self.t(q).inverse()


QuantumCircuit.t = t
QuantumCircuit.tdg = tdg
CompositeGate.t = t
CompositeGate.tdg = tdg



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Exception for errors raised by extensions module.
"""
from qiskit import QISKitError


class ExtensionError(QISKitError):
    """Base class for errors raised by extensions module."""

    def __init__(self, *message):
        """Set the error message."""
        self.message = ' '.join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Element of SU(2).
"""
from qiskit import Gate
from qiskit.extensions._extensionerror import ExtensionError
from qiskit import QuantumCircuit
from qiskit import CompositeGate


class UBase(Gate):
    """Element of SU(2)."""

    def __init__(self, param, qubit, circ=None):
        """Create new reset instruction."""
        if len(param) != 3:
            raise ExtensionError("expected 3 parameters")
        super(UBase, self).__init__("U", param, [qubit], circ)

    def qasm(self):
        """Return OPENQASM string."""
        theta = self.param[0]
        phi = self.param[1]
        lamb = self.param[2]
        qubit = self.arg[0]
        return self._qasmif("U(%.15f,%.15f,%.15f) %s[%d];" % (theta, phi,
                                                              lamb, qubit[0].name, qubit[1]))

    def inverse(self):
        """Invert this gate.

        U(theta,phi,lambda)^dagger = U(-theta,-lambda,-phi)
        """
        self.param[0] = -self.param[0]
        phi = self.param[1]
        self.param[1] = -self.param[2]
        self.param[2] = -phi
        return self

    def reapply(self, circ):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circ.ubase(self.arg[0]))


def u_base(self, tpl, q):
    """Apply U to q."""
    self._check_qubit(q)
    return self._attach(UBase(tpl, q, self))


QuantumCircuit.u_base = u_base
CompositeGate.u_base = u_base



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Backend for the unroller that prints OpenQASM.
"""
from ._backenderror import BackendError
from ._unrollerbackend import UnrollerBackend


class PrinterBackend(UnrollerBackend):
    """Backend for the unroller that prints OpenQASM.

    This backend also serves as an example class for other unroller backends.
    """

    def __init__(self, basis=None):
        """Setup this backend.

        basis is a list of operation name strings.
        """
        self.prec = 15
        self.creg = None
        self.cval = None
        self.gates = {}
        self.comments = False
        if basis:
            self.basis = basis
        else:
            self.basis = []
        self.listen = True
        self.in_gate = ""
        self.printed_gates = []

    def set_comments(self, comments):
        """Set comments to True to enable."""
        self.comments = comments

    def set_basis(self, basis):
        """Declare the set of user-defined gates to emit.

        basis is a list of operation name strings.
        """
        self.basis = basis

    def version(self, version):
        """Print the version string.

        v is a version number.
        """
        print("OPENQASM %s;" % version)

    def new_qreg(self, name, size):
        """Create a new quantum register.

        name = name of the register
        sz = size of the register
        """
        assert size >= 0, "invalid qreg size"
        print("qreg %s[%d];" % (name, size))

    def new_creg(self, name, size):
        """Create a new classical register.

        name = name of the register
        sz = size of the register
        """
        print("creg %s[%d];" % (name, size))

    def _gate_string(self, name):
        """Print OPENQASM for the named gate."""
        out = ""
        if self.gates[name]["opaque"]:
            out = "opaque " + name
        else:
            out = "gate " + name
        if self.gates[name]["n_args"] > 0:
            out += "(" + ",".join(self.gates[name]["args"]) + ")"
            out += " " + ",".join(self.gates[name]["bits"])
        if self.gates[name]["opaque"]:
            out += ";"
        else:
            out += "\n{\n" + self.gates[name]["body"].qasm() + "}"
        return out

    def define_gate(self, name, gatedata):
        """Define a new quantum gate.

        name is a string.
        gatedata is the AST node for the gate.
        """
        atomics = ["U", "CX", "measure", "reset", "barrier"]
        self.gates[name] = gatedata
        # Print out the gate definition if it is in self.basis
        if name in self.basis and name not in atomics:
            # Print the hierarchy of gates this gate calls
            if not self.gates[name]["opaque"]:
                calls = self.gates[name]["body"].calls()
                for call in calls:
                    if call not in self.printed_gates:
                        print(self._gate_string(call))
                        self.printed_gates.append(call)
            # Print the gate itself
            if name not in self.printed_gates:
                print(self._gate_string(name))
                self.printed_gates.append(name)

    def u(self, arg, qubit, nested_scope=None):
        """Fundamental single qubit gate.

        arg is 3-tuple of Node expression objects.
        qubit is (regname,idx) tuple.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        if self.listen:
            if "U" not in self.basis:
                self.basis.append("U")
            if self.creg is not None:
                print("if(%s==%d) " % (self.creg, self.cval), end="")
            print("U(%s,%s,%s) %s[%d];" % (arg[0].real(nested_scope),
                                           arg[1].real(nested_scope),
                                           arg[2].real(nested_scope),
                                           qubit[0],
                                           qubit[1]))

    def cx(self, qubit0, qubit1):
        """Fundamental two qubit gate.

        qubit0 is (regname,idx) tuple for the control qubit.
        qubit1 is (regname,idx) tuple for the target qubit.
        """
        if self.listen:
            if "CX" not in self.basis:
                self.basis.append("CX")
            if self.creg is not None:
                print("if(%s==%d) " % (self.creg, self.cval), end="")
            print("CX %s[%d],%s[%d];" % (qubit0[0], qubit0[1],
                                         qubit1[0], qubit1[1]))

    def measure(self, qubit, bit):
        """Measurement operation.

        qubit is (regname, idx) tuple for the input qubit.
        bit is (regname, idx) tuple for the output bit.
        """
        if "measure" not in self.basis:
            self.basis.append("measure")
        if self.creg is not None:
            print("if(%s==%d) " % (self.creg, self.cval), end="")
        print("measure %s[%d] -> %s[%d];" % (qubit[0], qubit[1],
                                             bit[0], bit[1]))

    def barrier(self, qubitlists):
        """Barrier instruction.

        qubitlists is a list of lists of (regname, idx) tuples.
        """
        if self.listen:
            if "barrier" not in self.basis:
                self.basis.append("barrier")
            names = []
            for qubitlist in qubitlists:
                if len(qubitlist) == 1:
                    names.append("%s[%d]" % (qubitlist[0][0], qubitlist[0][1]))
                else:
                    names.append("%s" % qubitlist[0][0])
            print("barrier %s;" % ",".join(names))

    def reset(self, qubit):
        """Reset instruction.

        qubit is a (regname, idx) tuple.
        """
        if "reset" not in self.basis:
            self.basis.append("reset")
        if self.creg is not None:
            print("if(%s==%d) " % (self.creg, self.cval), end="")
        print("reset %s[%d];" % (qubit[0], qubit[1]))

    def set_condition(self, creg, cval):
        """Attach a current condition.

        creg is a name string.
        cval is the integer value for the test.
        """
        self.creg = creg
        self.cval = cval
        if self.comments:
            print("// set condition %s, %s" % (creg, cval))

    def drop_condition(self):
        """Drop the current condition."""
        self.creg = None
        self.cval = None
        if self.comments:
            print("// drop condition")

    def start_gate(self, name, args, qubits, nested_scope=None):
        """Begin a custom gate.

        name is name string.
        args is list of Node expression objects.
        qubits is list of (regname, idx) tuples.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        if self.listen and self.comments:
            print("// start %s, %s, %s" % (name,
                                           list(map(lambda x:
                                                    str(x.real(nested_scope)),
                                                    args)),
                                           qubits))
        if self.listen and name not in self.basis \
           and self.gates[name]["opaque"]:
            raise BackendError("opaque gate %s not in basis" % name)
        if self.listen and name in self.basis:
            self.in_gate = name
            self.listen = False
            squbits = ["%s[%d]" % (x[0], x[1]) for x in qubits]
            if self.creg is not None:
                print("if(%s==%d) " % (self.creg, self.cval), end="")
            print(name, end="")
            if len(args) > 0:
                print("(%s)" % ",".join(map(lambda x:
                                            str(x.real(nested_scope)),
                                            args)), end="")
            print(" %s;" % ",".join(squbits))

    def end_gate(self, name, args, qubits, nested_scope=None):
        """End a custom gate.

        name is name string.
        args is list of Node expression objects.
        qubits is list of (regname, idx) tuples.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        if name == self.in_gate:
            self.in_gate = ""
            self.listen = True
        if self.listen and self.comments:
            print("// end %s, %s, %s" % (name,
                                         list(map(lambda x:
                                                  str(x.real(nested_scope)),
                                                  args)),
                                         qubits))

    def get_output(self):
        """This backend will return nothing, as the output has been directly
        written to screen"""
        pass



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Composite gate, a container for a sequence of unitary gates.
"""
from ._gate import Gate
from ._qiskiterror import QISKitError


class CompositeGate(Gate):
    """Composite gate, a sequence of unitary gates."""

    def __init__(self, name, param, args, circuit=None):
        """Create a new composite gate.

        name = instruction name string
        param = list of real parameters
        arg = list of pairs (Register, index)
        circ = QuantumCircuit or CompositeGate containing this gate
        """
        super().__init__(name, param, args, circuit)
        self.data = []  # gate sequence defining the composite unitary
        self.inverse_flag = False

    def has_register(self, register):
        """Test if this gate's circuit has the register r."""
        self.check_circuit()
        return self.circuit.has_register(register)

    def _modifiers(self, gate):
        """Apply any modifiers of this gate to another composite g."""
        if self.inverse_flag:
            gate.inverse()
        super()._modifiers(gate)

    def _attach(self, gate):
        """Attach a gate."""
        self.data.append(gate)
        return gate

    def _check_qubit(self, qubit):
        """Raise exception if q is not an argument or not qreg in circuit."""
        self.check_circuit()
        self.circuit._check_qubit(qubit)
        if (qubit[0].name, qubit[1]) not in map(
                lambda x: (x[0].name, x[1]), self.arg):
            raise QISKitError("qubit '%s[%d]' not argument of gate"
                                  % (qubit[0].name, qubit[1]))

    def _check_qreg(self, register):
        """Raise exception.

        if quantum register is not in this gate's circuit.
        """
        self.check_circuit()
        self.circuit._check_qreg(register)

    def _check_creg(self, register):
        """Raise exception.

        if classical register is not in this gate's circuit.
        """
        self.check_circuit()
        self.circuit._check_creg(register)

    def _check_dups(self, qubits):
        """Raise exception.

        if list of qubits contains duplicates.
        """
        squbits = set(qubits)
        if len(squbits) != len(qubits):
            raise QISKitError("duplicate qubit arguments")

    def qasm(self):
        """Return OPENQASM string."""
        return "\n".join([g.qasm() for g in self.data])

    def inverse(self):
        """Invert this gate."""
        self.data = [gate.inverse() for gate in reversed(self.data)]
        self.inverse_flag = not self.inverse_flag
        return self

    def q_if(self, *qregs):
        """Add controls to this gate."""
        self.data = [gate.q_if(qregs) for gate in self.data]
        return self

    def c_if(self, c, val):
        """Add classical control register."""
        self.data = [gate.c_if(c, val) for gate in self.data]
        return self



"""Interface to a fast C++ QASM simulator.

Authors: Erick Winston <ewinston@us.ibm.com>
         Christopher J. Wood <cjwood@us.ibm.com>

"""

import json
import subprocess
from subprocess import PIPE, CalledProcessError
import numpy as np
from ._simulatorerror import SimulatorError

__configuration = {
    "name": "local_qasm_cpp_simulator",
    "url": "https://github.com/IBM/qiskit-sdk-py",
    "simulator": True,
    "description": "A c++ simulator for qasm files",
    "coupling_map": "all-to-all",
    "basis_gates": "u1,u2,u3,cx,id"
}


class QasmCppSimulator:
    """
    Interface to a fast C++ QASM simulator.
    """

    def __init__(self, job):
        # check for config file
        if 'config' in job:
            self.config = job['config']
        else:
            self.config = {}
        self.circuit = {'qasm': json.loads(job['compiled_circuit'].decode()),
                        'config': self.config}
        self.result = {}
        self.result['data'] = {}
        self._shots = job['config']['shots']
        self._seed = job['config']['seed']
        # Number of threads for simulator
        if 'threads' in self.config:
            self._threads = self.config['threads']
        else:
            self._threads = 1
        # Location of simulator exe
        if 'exe' in self.config:
            self._exe = self.config['exe']
        else:
            self._exe = 'qasm_simulator'
        # C++ simulator backend
        if 'simulator' in self.config:
            self._cpp_backend = self.config['simulator']
        else:
            self._cpp_backend = 'qubit'
        self._number_of_operations = len(self.circuit['qasm']['operations'])
        # This assumes we are getting a quick return help message.
        # so _localsimulator can quickly determine whether the compiled
        # simulator is available.
        try:
            subprocess.check_output([self._exe], stderr=subprocess.STDOUT)
        except CalledProcessError:
            pass
        except FileNotFoundError:
            try:
                subprocess.check_output(
                    ['./' + self._exe], stderr=subprocess.STDOUT)
            except CalledProcessError:
                # simulator with no arguments returns 1
                # so this is the "success" case
                self._exe = './' + self._exe
            except FileNotFoundError:
                cmd = '"{0}" or "{1}" '.format(self._exe, './' + self._exe)
                raise FileNotFoundError(cmd)

    def run(self, silent=True):
        """
        Run simulation on C++ simulator.
        """
        # Build the command line execution string
        cmd = self._exe + ' -i - -c - -f qiskit'
        cmd += ' -n {shots:d}'.format(shots=self._shots)
        cmd += ' -t {threads:d}'.format(threads=self._threads)
        cmd += ' -b {backend:s}'.format(backend=self._cpp_backend)
        if self._seed is not None:
            if self._seed >= 0:
                if isinstance(self._seed, float):
                    # _quantumprogram.py usually generates float in [0,1]
                    # try to convert to integer which C++ random expects.
                    self._seed = hash(self._seed)
                cmd += ' -s {seed:d}'.format(seed=self._seed)
            else:
                raise TypeError('seed needs to be an unsigned integer')
        # Open subprocess and execute external command
        with subprocess.Popen(cmd.split(),
                              stdin=PIPE, stdout=PIPE, stderr=PIPE) as proc:
            cin = json.dumps(self.circuit).encode()
            cout, cerr = proc.communicate(cin)
        if len(cerr) == 0:
            # no error messages, load std::cout
            cresult = json.loads(cout.decode())
            # convert possible complex valued result fields
            for k in ['state', 'saved_states', 'inner_products']:
                parse_complex(cresult['data'], k)
        else:
            # custom "backend" or "result" exception handler here?
            raise SimulatorError('local_qasm_cpp_simulator returned: {0}\n{1}'.
                            format(cout.decode(), cerr.decode()))
        # Add simulator data
        self.result['data'] = cresult['data']
        # Add simulation time (in seconds)
        self.result['time_taken'] = cresult['time_taken']
        # Report finished
        self.result['status'] = 'DONE'
        return self.result


def parse_complex(output, key):
    """
    Parse complex entries in C++ simulator output.

    This function converts complex numbers in the C++ simulator output
    into python complex numbers. In JSON c++ output complex entries are
    formatted as::

        z = [re(z), im(z)]
        vec = [re(vec), im(vec)]
        ket = {'00':[re(v[00]), im(v[00])], '01': etc...}

    Args:
        output (dict): simulator output.
        key (str): the output key to search for.
    """
    if key in output:
        ref = output[key]
        if isinstance(ref, list):
            if isinstance(ref[0], list):
                # convert complex vector
                for i, j in enumerate(ref):
                    ref[i] = np.array(j[0]) + 1j * np.array(j[1])
            elif isinstance(ref[0], dict):
                # convert complex ket-form
                for i, j in enumerate(ref):
                    for k in j.keys():
                        ref[i][k] = j[k][0] + 1j * j[k][1]
            elif len(ref) == 2:
                # convert complex scalar
                ref = ref[0] + 1j * ref[1]



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Backend for the unroller that creates a DAGCircuit object.
"""
from ._unrollerbackend import UnrollerBackend
from ._backenderror import BackendError
from ..dagcircuit import DAGCircuit


class DAGBackend(UnrollerBackend):
    """Backend for the unroller that creates a DAGCircuit object."""

    def __init__(self, basis=None):
        """Setup this backend.

        basis is a list of operation name strings.
        """
        self.prec = 15
        self.creg = None
        self.cval = None
        self.circuit = DAGCircuit()
        if basis:
            self.basis = basis
        else:
            self.basis = []
        self.listen = True
        self.in_gate = ""
        self.gates = {}

    def set_basis(self, basis):
        """Declare the set of user-defined gates to emit."""
        self.basis = basis

    def define_gate(self, name, gatedata):
        """Record and pass down the data for this gate."""
        self.gates[name] = gatedata
        self.circuit.add_gate_data(name, gatedata)

    def version(self, version):
        """Accept the version string.

        v is a version number.
        """
        pass

    def new_qreg(self, name, size):
        """Create a new quantum register.

        name = name of the register
        sz = size of the register
        """
        self.circuit.add_qreg(name, size)

    def new_creg(self, name, size):
        """Create a new classical register.

        name = name of the register
        sz = size of the register
        """
        self.circuit.add_creg(name, size)

    def u(self, arg, qubit, nested_scope=None):
        """Fundamental single qubit gate.

        arg is 3-tuple of Node expression objects.
        qubit is (regname,idx) tuple.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        if self.listen:
            if self.creg is not None:
                condition = (self.creg, self.cval)
            else:
                condition = None
            if "U" not in self.basis:
                self.basis.append("U")
                self.circuit.add_basis_element("U", 1, 0, 3)
            self.circuit.apply_operation_back(
                "U", [qubit], [], list(map(lambda x: x.real(nested_scope),
                                           arg)), condition)

    def cx(self, qubit0, qubit1):
        """Fundamental two-qubit gate.

        qubit0 is (regname, idx) tuple for the control qubit.
        qubit1 is (regname, idx) tuple for the target qubit.
        """
        if self.listen:
            if self.creg is not None:
                condition = (self.creg, self.cval)
            else:
                condition = None
            if "CX" not in self.basis:
                self.basis.append("CX")
                self.circuit.add_basis_element("CX", 2)
            self.circuit.apply_operation_back("CX", [qubit0, qubit1], [],
                                              [], condition)

    def measure(self, qubit, bit):
        """Measurement operation.

        qubit is (regname, idx) tuple for the input qubit.
        bit is (regname, idx) tuple for the output bit.
        """
        if self.creg is not None:
            condition = (self.creg, self.cval)
        else:
            condition = None
        if "measure" not in self.basis:
            self.basis.append("measure")
            self.circuit.add_basis_element("measure", 1, 1)
        self.circuit.apply_operation_back(
            "measure", [qubit], [bit], [], condition)

    def barrier(self, qubitlists):
        """Barrier instruction.

        qubitlists is a list of lists of (regname, idx) tuples.
        """
        if self.listen:
            names = []
            for x in qubitlists:
                for j in range(len(x)):
                    names.append(x[j])
            if "barrier" not in self.basis:
                self.basis.append("barrier")
                self.circuit.add_basis_element("barrier", -1)
            self.circuit.apply_operation_back("barrier", names)

    def reset(self, qubit):
        """Reset instruction.

        qubit is a (regname, idx) tuple.
        """
        if self.creg is not None:
            condition = (self.creg, self.cval)
        else:
            condition = None
        if "reset" not in self.basis:
            self.basis.append("reset")
            self.circuit.add_basis_element("reset", 1)
        self.circuit.apply_operation_back("reset", [qubit], [], [], condition)

    def set_condition(self, creg, cval):
        """Attach a current condition.

        creg is a name string.
        cval is the integer value for the test.
        """
        self.creg = creg
        self.cval = cval

    def drop_condition(self):
        """Drop the current condition."""
        self.creg = None
        self.cval = None

    def start_gate(self, name, args, qubits, nested_scope=None):
        """Begin a custom gate.

        name is name string.
        args is list of Node expression objects.
        qubits is list of (regname, idx) tuples.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        if self.listen and name not in self.basis \
           and self.gates[name]["opaque"]:
            raise BackendError("opaque gate %s not in basis" % name)
        if self.listen and name in self.basis:
            if self.creg is not None:
                condition = (self.creg, self.cval)
            else:
                condition = None
            self.in_gate = name
            self.listen = False
            self.circuit.add_basis_element(name, len(qubits), 0, len(args))
            self.circuit.apply_operation_back(
                name, qubits, [], list(map(lambda x: x.real(nested_scope),
                                           args)), condition)

    def end_gate(self, name, args, qubits, nested_scope=None):
        """End a custom gate.

        name is name string.
        args is list of Node expression objects.
        qubits is list of (regname, idx) tuples.
        nested_scope is a list of dictionaries mapping expression variables
        to Node expression objects in order of increasing nesting depth.
        """
        if name == self.in_gate:
            self.in_gate = ""
            self.listen = True

    def get_output(self):
        """Returns the generated circuit."""
        return self.circuit



# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Directed graph object for representing coupling between qubits.

The nodes of the graph correspond to named qubits and the directed edges
indicate which qubits are coupled and the permitted direction of CNOT gates.
The object has a distance function that can be used to map quantum circuits
onto a device with this coupling.
"""
from collections import OrderedDict
import networkx as nx
from ._couplingerror import CouplingError


def coupling_dict2list(couplingdict):
    """Convert coupling map dictionary into list.

    Example dictionary format: {0: [1, 2], 1: [2]}
    Example list format: [[0, 1], [0, 2], [1, 2]]

    We do not do any checking of the input.

    Return coupling map in list format.
    """
    if not couplingdict:
        return None
    couplinglist = []
    for ctl, tgtlist in couplingdict.items():
        for tgt in tgtlist:
            couplinglist.append([ctl, tgt])
    return couplinglist


def coupling_list2dict(couplinglist):
    """Convert coupling map list into dictionary.

    Example list format: [[0, 1], [0, 2], [1, 2]]
    Example dictionary format: {0: [1, 2], 1: [2]}

    We do not do any checking of the input.

    Return coupling map in dict format.
    """
    if not couplinglist:
        return None
    couplingdict = {}
    for pair in couplinglist:
        if pair[0] in couplingdict:
            couplingdict[pair[0]].append(pair[1])
        else:
            couplingdict[pair[0]] = [pair[1]]
    return couplingdict


class Coupling:
    """
    Directed graph specifying fixed coupling.

    Nodes correspond to qubits and directed edges correspond to permitted
    CNOT gates
    """

    def __init__(self, couplingdict=None):
        """
        Create coupling graph.

        By default, the coupling graph has no nodes. The optional couplingdict
        specifies the graph as an adjacency list. For example,
        couplingdict = {0: [1, 2], 1: [2]}.
        """
        # self.qubits is dict from qubit (regname,idx) tuples to node indices
        self.qubits = OrderedDict()
        # self.index_to_qubit is a dict from node indices to qubits
        self.index_to_qubit = {}
        # self.node_counter is integer counter for labeling nodes
        self.node_counter = 0
        # self.G is the coupling digraph
        self.G = nx.DiGraph()
        # self.dist is a dict of dicts from node pairs to distances
        # it must be computed, it is the distance on the digraph
        self.dist = None
        # Add edges to the graph if the couplingdict is present
        if couplingdict is not None:
            for v0, alist in couplingdict.items():
                for v1 in alist:
                    regname = "q"
                    self.add_edge((regname, v0), (regname, v1))
            self.compute_distance()

    def size(self):
        """Return the number of qubits in this graph."""
        return len(self.qubits)

    def get_qubits(self):
        """Return the qubits in this graph as (qreg, index) tuples."""
        return list(self.qubits.keys())

    def get_edges(self):
        """Return a list of edges in the coupling graph.

        Each edge is a pair of qubits and each qubit is a tuple (qreg, index).
        """
        return list(map(lambda x: (self.index_to_qubit[x[0]],
                                   self.index_to_qubit[x[1]]), self.G.edges()))

    def add_qubit(self, name):
        """
        Add a qubit to the coupling graph.

        name = tuple (regname, idx) for qubit
        """
        if name in self.qubits:
            raise CouplingError("%s already in coupling graph" % name)

        self.node_counter += 1
        self.G.add_node(self.node_counter)
        self.G.node[self.node_counter]["name"] = name
        self.qubits[name] = self.node_counter
        self.index_to_qubit[self.node_counter] = name

    def add_edge(self, s_name, d_name):
        """
        Add directed edge to coupling graph.

        s_name = source qubit tuple
        d_name = destination qubit tuple
        """
        if s_name not in self.qubits:
            self.add_qubit(s_name)
        if d_name not in self.qubits:
            self.add_qubit(d_name)
        self.G.add_edge(self.qubits[s_name], self.qubits[d_name])

    def connected(self):
        """
        Test if the graph is connected.

        Return True if connected, False otherwise
        """
        return nx.is_weakly_connected(self.G)

    def compute_distance(self):
        """
        Compute the distance function on pairs of nodes.

        The distance map self.dist is computed from the graph using
        all_pairs_shortest_path_length.
        """
        if not self.connected():
            raise CouplingError("coupling graph not connected")
        lengths = nx.all_pairs_shortest_path_length(self.G.to_undirected())
        self.dist = {}
        for i in self.qubits.keys():
            self.dist[i] = {}
            for j in self.qubits.keys():
                self.dist[i][j] = lengths[self.qubits[i]][self.qubits[j]]

    def distance(self, q1, q2):
        """Return the distance between qubit q1 to qubit q2."""
        if self.dist is None:
            raise CouplingError("distance has not been computed")
        if q1 not in self.qubits:
            raise CouplingError("%s not in coupling graph" % q1)
        if q2 not in self.qubits:
            raise CouplingError("%s not in coupling graph" % q2)
        return self.dist[q1][q2]

    def __str__(self):
        """Return a string representation of the coupling graph."""
        s = "qubits: "
        s += ", ".join(["%s[%d] @ %d" % (k[0], k[1], v)
                        for k, v in self.qubits.items()])
        s += "\nedges: "
        s += ", ".join(["%s[%d]-%s[%d]" % (e[0][0], e[0][1], e[1][0], e[1][1])
                        for e in self.get_edges()])
        return s



# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
Qasm Program Class
"""
# pylint: disable=line-too-long
import time
import random
import json
import os
import string
import re
import copy

# use the external IBMQuantumExperience Library
from IBMQuantumExperience.IBMQuantumExperience import IBMQuantumExperience

# Stable Modules
from . import QuantumRegister
from . import ClassicalRegister
from . import QuantumCircuit
from . import QISKitError

# Beta Modules
from . import unroll
from . import qasm
from . import mapper

# Local Simulator Modules
from . import simulators

import sys
sys.path.append("..")
import qiskit.extensions.standard

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


def convert(name):
    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()


class QuantumProgram(object):
    """Quantum Program Class.

     Class internal properties.

     Elements that are not python identifiers or string constants are denoted
     by "--description (type)--". For example, a circuit's name is denoted by
     "--circuit name (string)--" and might have the value "teleport".

     Internal::

        __quantum_registers (list[dic]): An dictionary of quantum registers
            used in the quantum program.
            __quantum_registers =
                {
                --register name (string)--: QuantumRegistor,
                }
        __classical_registers (list[dic]): An ordered list of classical
            registers used in the quantum program.
            __classical_registers =
                {
                --register name (string)--: ClassicalRegistor,
                }
        __quantum_program (dic): An dictionary of quantum circuits
            __quantum_program =
                {
                --circuit name (string)--:  --circuit object --,
                }
        __init_circuit (obj): A quantum circuit object for the initial quantum
            circuit
        __ONLINE_BACKENDS (list[str]): A list of online backends
        __LOCAL_BACKENDS (list[str]): A list of local backends
     """
    # -- FUTURE IMPROVEMENTS --
    # TODO: for status results make ALL_CAPS (check) or some unified method
    # TODO: Jay: coupling_map, basis_gates will move into a config object
    # only exists once you set the api to use the online backends
    __api = {}
    __api_config = {}

    def __init__(self, specs=None):
        self.__quantum_registers = {}
        self.__classical_registers = {}
        self.__quantum_program = {}  # stores all the quantum programs
        self.__init_circuit = None  # stores the intial quantum circuit of the
        # program
        self.__ONLINE_BACKENDS = []
        self.__LOCAL_BACKENDS = self.local_backends()
        self.mapper = mapper
        if specs:
            self.__init_specs(specs)

    ###############################################################
    # methods to initiate an build a quantum program
    ###############################################################

    def __init_specs(self, specs, verbose=False):
        """Populate the Quantum Program Object with initial Specs.

        Args:
            specs (dict):
                    Q_SPECS = {
                        "circuits": [{
                            "name": "Circuit",
                            "quantum_registers": [{
                                "name": "qr",
                                "size": 4
                            }],
                            "classical_registers": [{
                                "name": "cr",
                                "size": 4
                            }]
                        }],
            verbose (bool): controls how information is returned.

        Returns:
            Sets up a quantum circuit.
        """
        quantumr = []
        classicalr = []
        if "circuits" in specs:
            for circuit in specs["circuits"]:
                quantumr = self.create_quantum_registers(
                    circuit["quantum_registers"])
                classicalr = self.create_classical_registers(
                    circuit["classical_registers"])
                self.create_circuit(name=circuit["name"], qregisters=quantumr,
                                    cregisters=classicalr)
        # TODO: Jay: I think we should return function handles for the
        # registers and circuit. So that we dont need to get them after we
        # create them with get_quantum_register etc

    def create_quantum_register(self, name, size, verbose=False):
        """Create a new Quantum Register.

        Args:
            name (str): the name of the quantum register
            size (int): the size of the quantum register
            verbose (bool): controls how information is returned.

        Returns:
            internal reference to a quantum register in __quantum_register s
        """
        if name in self.__quantum_registers:
            if size != len(self.__quantum_registers[name]):
                raise QISKitError("Can't make this register: Already in"
                                  " program with different size")
            if verbose is True:
                print(">> quantum_register exists:", name, size)
        else:
            if verbose is True:
                print(">> new quantum_register created:", name, size)
            self.__quantum_registers[name] = QuantumRegister(name, size)
        return self.__quantum_registers[name]

    def destroy_quantum_register(self, name, verbose=False):
        """Destroy an existing Quantum Register.

        Args:
            name (str): the name of the quantum register

        Raises:
            QISKitError: if the register does not exist in the program.
        """
        if name not in self.__quantum_registers:
            raise QISKitError("Can't destroy this register: Not present")
        else:
            if verbose:
                print(">> quantum_register destroyed: %s", name)
            del self.__quantum_registers[name]

    def create_quantum_registers(self, register_array):
        """Create a new set of Quantum Registers based on a array of them.

        Args:
            register_array (list[dict]): An array of quantum registers in
                dictionay format::

                    "quantum_registers": [
                        {
                        "name": "qr",
                        "size": 4
                        },
                        ...
                    ]
        Returns:
            Array of quantum registers objects
        """
        new_registers = []
        for register in register_array:
            register = self.create_quantum_register(
                register["name"], register["size"])
            new_registers.append(register)
        return new_registers

    def destroy_quantum_registers(self, register_array):
        """Destroy a set of Quantum Registers based on a array of them.

        Args:
            register_array (list[dict]): An array of quantum registers in
                dictionay format::

                    "quantum_registers": [
                        {
                        "name": "qr",
                        },
                        ...
                    ]
                "size" may be a key for compatibility, but is ignored.
        """
        for register in register_array:
            self.destroy_quantum_register(register["name"])

    def create_classical_register(self, name, size, verbose=False):
        """Create a new Classical Register.

        Args:
            name (str): the name of the quantum register
            size (int): the size of the quantum register
            verbose (bool): controls how information is returned.
        Returns:
            internal reference to a quantum register in __quantum_register
        """
        if name in self.__classical_registers:
            if size != len(self.__classical_registers[name]):
                raise QISKitError("Can't make this register: Already in"
                                  " program with different size")
            if verbose is True:
                print(">> classical register exists:", name, size)
        else:
            if verbose is True:
                print(">> new classical register created:", name, size)
            self.__classical_registers[name] = ClassicalRegister(name, size)
        return self.__classical_registers[name]

    def create_classical_registers(self, registers_array):
        """Create a new set of Classical Registers based on a array of them.

        Args:
            register_array (list[dict]): An array of classical registers in
                dictionay fromat::

                    "classical_registers": [
                        {
                        "name": "qr",
                        "size": 4
                        },
                        ...
                    ]
        Returns:
            Array of clasical registers objects
        """
        new_registers = []
        for register in registers_array:
            new_registers.append(self.create_classical_register(
                register["name"], register["size"]))
        return new_registers

    def destroy_classical_register(self, name, verbose=False):
        """Destroy an existing Classical Register.

        Args:
            name (str): the name of the classical register

        Raises:
            QISKitError: if the register does not exist in the program.
        """
        if name not in self.__classical_registers:
            raise QISKitError("Can't destroy this register: Not present")
        else:
            if verbose:
                print(">> classical register destroyed: %s", name)
            del self.__classical_registers[name]

    def destroy_classical_registers(self, registers_array):
        """Destroy a set of Classical Registers based on a array of them.

        Args:
            registers_array (list[dict]): An array of classical registers in
                dictionay fromat::

                    "classical_registers": [
                        {
                        "name": "qr",
                        },
                        ...
                    ]
                "size" may be a key for compatibility, but is ignored.
        """
        for register in registers_array:
            self.destroy_classical_register(register["name"])

    def create_circuit(self, name, qregisters=None, cregisters=None):
        """Create a empty Quantum Circuit in the Quantum Program.

        Args:
            name (str): the name of the circuit
            qregisters list(object): is an Array of Quantum Registers by object
                reference
            cregisters list(object): is an Array of Classical Registers by
                object reference

        Returns:
            A quantum circuit is created and added to the Quantum Program
        """
        if not qregisters:
            qregisters = []
        if not cregisters:
            cregisters = []
        quantum_circuit = QuantumCircuit()
        if not self.__init_circuit:
            self.__init_circuit = quantum_circuit
        for register in qregisters:
            quantum_circuit.add(register)
        for register in cregisters:
            quantum_circuit.add(register)
        self.add_circuit(name, quantum_circuit)
        return self.__quantum_program[name]

    def destroy_circuit(self, name):
        """Destroy a Quantum Circuit in the Quantum Program. This will not
        destroy any registers associated with the circuit.

        Args:
            name (str): the name of the circuit

        Raises:
            QISKitError: if the register does not exist in the program.
        """
        if name not in self.__quantum_program:
            raise QISKitError("Can't destroy this circuit: Not present")
        del self.__quantum_program[name]

    def add_circuit(self, name, quantum_circuit):
        """Add a new circuit based on an Object representation.

        Args:
            name (str): the name of the circuit to add.
            quantum_circuit: a quantum circuit to add to the program-name
        Returns:
            the quantum circuit is added to the object.
        """
        for qname, qreg in quantum_circuit.get_qregs().items():
            self.create_quantum_register(qname, len(qreg))
        for cname, creg in quantum_circuit.get_cregs().items():
            self.create_classical_register(cname, len(creg))
        self.__quantum_program[name] = quantum_circuit

    def load_qasm_file(self, qasm_file, name=None, verbose=False):
        """ Load qasm file into the quantum program.

        Args:
            qasm_file (str): a string for the filename including its location.
            name (str or None, optional): the name of the quantum circuit after
                loading qasm text into it. If no name is give the name is of
                the text file.
            verbose (bool, optional): controls how information is returned.
        Retuns:
            Adds a quantum circuit with the gates given in the qasm file to the
            quantum program and returns the name to be used to get this circuit
        """
        if not os.path.exists(qasm_file):
            raise QISKitError('qasm file "{0}" not found'.format(qasm_file))
        if not name:
            name = os.path.splitext(os.path.basename(qasm_file))[0]
        node_circuit = qasm.Qasm(filename=qasm_file).parse()  # Node (AST)
        if verbose is True:
            print("circuit name: " + name)
            print("******************************")
            print(node_circuit.qasm())
        # current method to turn it a DAG quantum circuit.
        basis_gates = "u1,u2,u3,cx,id"  # QE target basis
        unrolled_circuit = unroll.Unroller(node_circuit,
                                           unroll.CircuitBackend(basis_gates.split(",")))
        circuit_unrolled = unrolled_circuit.execute()
        self.add_circuit(name, circuit_unrolled)
        return name

    def load_qasm_text(self, qasm_string, name=None, verbose=False):
        """ Load qasm string in the quantum program.

        Args:
            qasm_string (str): a string for the file name.
            name (str): the name of the quantum circuit after loading qasm
                text into it. If no name is give the name is of the text file.
            verbose (bool): controls how information is returned.
        Retuns:
            Adds a quantum circuit with the gates given in the qasm string to
            the quantum program.
        """
        node_circuit = qasm.Qasm(data=qasm_string).parse()  # Node (AST)
        if not name:
            # Get a random name if none is give
            name = "".join([random.choice(string.ascii_letters+string.digits)
                            for n in range(10)])
        if verbose is True:
            print("circuit name: " + name)
            print("******************************")
            print(node_circuit.qasm())
        # current method to turn it a DAG quantum circuit.
        basis_gates = "u1,u2,u3,cx,id"  # QE target basis
        unrolled_circuit = unroll.Unroller(node_circuit,
                                           unroll.CircuitBackend(basis_gates.split(",")))
        circuit_unrolled = unrolled_circuit.execute()
        self.add_circuit(name, circuit_unrolled)
        return name

    ###############################################################
    # methods to get elements from a QuantumProgram
    ###############################################################

    def get_quantum_register(self, name):
        """Return a Quantum Register by name.

        Args:
            name (str): the name of the quantum circuit
        Returns:
            The quantum registers with this name
        """
        try:
            return self.__quantum_registers[name]
        except KeyError:
            raise KeyError('No quantum register "{0}"'.format(name))

    def get_classical_register(self, name):
        """Return a Classical Register by name.

        Args:
            name (str): the name of the quantum circuit
        Returns:
            The classical registers with this name
        """
        try:
            return self.__classical_registers[name]
        except KeyError:
            raise KeyError('No classical register "{0}"'.format(name))

    def get_quantum_register_names(self):
        """Return all the names of the quantum Registers."""
        return self.__quantum_registers.keys()

    def get_classical_register_names(self):
        """Return all the names of the classical Registers."""
        return self.__classical_registers.keys()

    def get_circuit(self, name):
        """Return a Circuit Object by name
        Args:
            name (str): the name of the quantum circuit
        Returns:
            The quantum circuit with this name
        """
        try:
            return self.__quantum_program[name]
        except KeyError:
            raise KeyError('No quantum circuit "{0}"'.format(name))

    def get_circuit_names(self):
        """Return all the names of the quantum circuits."""
        return self.__quantum_program.keys()

    def get_qasm(self, name):
        """Get qasm format of circuit by name.

        Args:
            name (str): name of the circuit

        Returns:
            The quantum circuit in qasm format
        """
        quantum_circuit = self.get_circuit(name)
        return quantum_circuit.qasm()

    def get_qasms(self, list_circuit_name):
        """Get qasm format of circuit by list of names.

        Args:
            list_circuit_name (list[str]): names of the circuit

        Returns:
            List of quantum circuit in qasm format
        """
        qasm_source = []
        for name in list_circuit_name:
            qasm_source.append(self.get_qasm(name))
        return qasm_source

    def get_initial_circuit(self):
        """Return the initialization Circuit."""
        return self.__init_circuit

    ###############################################################
    # methods for working with backends
    ###############################################################

    def set_api(self, token, url, hub=None, group=None, project=None,
                proxies=None, verify=True):
        """ Setup the API.

        Does not catch exceptions from IBMQuantumExperience.

        Args:
            token (str): The token used to register on the online backend such
                as the quantum experience.
            url (str): The url used for online backend such as the quantum
                experience.
            hub (str): The hub used for online backend.
            group (str): The group used for online backend.
            project (str): The project used for online backend.
            proxies (dict): Proxy configuration for the API, as a dict with
                'urls' and credential keys.
            verify (bool): If False, ignores SSL certificates errors.
        Returns:
            Nothing but fills the __ONLINE_BACKENDS, __api, and __api_config
        Raises:
              ConnectionError: if the API instantiation failed.
        """
        try:

            config_dict = {
                'url': url,
                'hub': hub,
                'group': group,
                'project': project
            }
            if proxies:
                config_dict['proxies'] = proxies
            self.__api = IBMQuantumExperience(token, config_dict, verify)
        except Exception as ex:
            root_exception = ex
            if 'License required' in str(ex):
                # For the 401 License required exception from the API, be
                # less verbose with the exceptions.
                root_exception = None
            raise ConnectionError("Couldn't connect to IBMQuantumExperience server: {0}"
                                  .format(ex)) from root_exception

        self.__ONLINE_BACKENDS = self.online_backends()
        self.__api_config["token"] = token
        self.__api_config["url"] = {"url": url}
        self.__api_config["config"] = config_dict.copy()

    def set_api_hubs_config(self, hub, group, project):
        """Update the API hubs configuration, replacing the previous one.

         hub (str): The hub used for online backend.
         group (str): The group used for online backend.
         project (str): The project used for online backend.
        """
        config_dict = {
            'hub': hub,
            'group': group,
            'project': project
        }

        for k, v in config_dict.items():
            self.__api.config[k] = v
            self.__api_config['config'][k] = v

    def get_api_config(self):
        """Return the program specs."""
        return self.__api_config

    def get_api(self):
        """Returns a function handle to the API."""
        return self.__api

    def save(self, file_name=None, beauty=False):
        """ Save Quantum Program in a Json file.

        Args:
            file_name (str): file name and path.
            beauty (boolean): save the text with indent 4 to make it readable.

        Returns:
            The dictionary with the status and result of the operation

        Raises:
            When you don't provide a correct file name
                raise a LookupError.
            When something happen with the file management
                raise a LookupError.
        """
        if file_name is None:
            error = {"status": "Error", "result": "Not filename provided"}
            raise LookupError(error['result'])

        if beauty:
            indent = 4
        else:
            indent = 0

        elemements_to_save = self.__quantum_program
        elements_saved = {}

        for circuit in elemements_to_save:
            elements_saved[circuit] = {}
            elements_saved[circuit]["qasm"] = elemements_to_save[circuit].qasm()

        try:
            with open(file_name, 'w') as save_file:
                json.dump(elements_saved, save_file, indent=indent)
            return {'status': 'Done', 'result': elemements_to_save}
        except ValueError:
            error = {'status': 'Error', 'result': 'Some Problem happened to save the file'}
            raise LookupError(error['result'])

    def load(self, file_name=None):
        """ Load Quantum Program Json file into the Quantum Program object.

        Args:
            file_name (str): file name and path.

        Returns:
            The dictionary with the status and result of the operation

        Raises:
            When you don't provide a correct file name
                raise a LookupError.
            When something happen with the file management
                raise a LookupError.
        """
        if file_name is None:
            error = {"status": "Error", "result": "Not filename provided"}
            raise LookupError(error['result'])

        elemements_to_load = {}

        try:
            with open(file_name, 'r') as load_file:
                elemements_loaded = json.load(load_file)

            for circuit in elemements_loaded:
                circuit_qasm = elemements_loaded[circuit]["qasm"]
                elemements_loaded[circuit] = qasm.Qasm(data=circuit_qasm).parse()
            self.__quantum_program = elemements_loaded

            return {"status": 'Done', 'result': self.__quantum_program}

        except ValueError:
            error = {'status': 'Error', 'result': 'Some Problem happened to load the file'}
            raise LookupError(error['result'])

    def available_backends(self):
        """All the backends that are seen by QISKIT."""
        return self.__ONLINE_BACKENDS + self.__LOCAL_BACKENDS

    def local_backends(self):
        """Get the local backends."""
        return simulators._localsimulator.local_backends()

    def online_backends(self):
        """Get the online backends.

        Queries network API if it exists and gets the backends that are online.

        Returns:
            List of online backends if the online api has been set or an empty
            list of it has not been set.
        """
        if self.get_api():
            return [backend['name'] for backend in self.__api.available_backends()]
        else:
            return []

    def online_simulators(self):
        """Gets online simulators via QX API calls.

        Returns:
            List of online simulator names.
        """
        simulators = []
        if self.get_api():
            for backend in self.__api.available_backends():
                if backend['simulator']:
                    simulators.append(backend['name'])
        return simulators

    def online_devices(self):
        """Gets online devices via QX API calls.

        Returns:
            List of online simulator names.
        """
        devices = []
        if self.get_api():
            for backend in self.__api.available_backends():
                if not backend['simulator']:
                    devices.append(backend['name'])
        return devices

    def get_backend_status(self, backend):
        """Return the online backend status.

        It uses QX API call or by local backend is the name of the
        local or online simulator or experiment.

        Args:
            banckend (str): The backend to check
        """

        if backend in self.__ONLINE_BACKENDS:
            return self.__api.backend_status(backend)
        elif backend in self.__LOCAL_BACKENDS:
            return {'available': True}
        else:
            err_str = 'the backend "{0}" is not available'.format(backend)
            raise ValueError(err_str)

    def get_backend_configuration(self, backend, list_format=False):
        """Return the configuration of the backend.

        The return is via QX API call.

        Args:
            backend (str):  Name of the backend.

        Returns:
            The configuration of the named backend.

        Raises:
            If a configuration for the named backend can't be found
            raise a LookupError.
        """
        if self.get_api():
            configuration_edit = {}
            for configuration in self.__api.available_backends():
                if configuration['name'] == backend:
                    for key in configuration:
                        new_key = convert(key)
                        # TODO: removed these from the API code
                        if new_key not in ['id', 'serial_number', 'topology_id',
                                           'status', 'coupling_map']:
                            configuration_edit[new_key] = configuration[key]
                        if new_key == 'coupling_map':
                            if configuration[key] == 'all-to-all':
                                configuration_edit[new_key] = \
                                    configuration[key]
                            else:
                                if not list_format:
                                    cmap = mapper.coupling_list2dict(
                                                configuration[key])
                                else:
                                    cmap = configuration[key]
                                configuration_edit[new_key] = cmap
                    return configuration_edit
        for configuration in simulators.local_configuration:
            if configuration['name'] == backend:
                return configuration
        raise LookupError(
            'backend configuration for "{0}" not found'.format(backend))

    def get_backend_calibration(self, backend):
        """Return the online backend calibrations.

        The return is via QX API call.

        Args:
            backend (str):  Name of the backend.

        Returns:
            The configuration of the named backend.

        Raises:
            If a configuration for the named backend can't be found
            raise a LookupError.
        """
        if backend in self.__ONLINE_BACKENDS:
            calibrations = self.__api.backend_calibration(backend)
            calibrations_edit = {}
            for key, vals in calibrations.items():
                new_key = convert(key)
                calibrations_edit[new_key] = vals
            return calibrations_edit
        elif  backend in self.__LOCAL_BACKENDS:
            return {'backend': backend, 'calibrations': None}
        else:
            raise LookupError(
                'backend calibration for "{0}" not found'.format(backend))

    def get_backend_parameters(self, backend):
        """Return the online backend parameters.

        The return is via QX API call.

        Args:
            backend (str):  Name of the backend.

        Returns:
            The configuration of the named backend.

        Raises:
            If a configuration for the named backend can't be found
            raise a LookupError.
        """
        if backend in self.__ONLINE_BACKENDS:
            parameters = self.__api.backend_parameters(backend)
            parameters_edit = {}
            for key, vals in parameters.items():
                new_key = convert(key)
                parameters_edit[new_key] = vals
            return parameters_edit
        elif backend in self.__LOCAL_BACKENDS:
            return {'backend': backend, 'parameters': None}
        else:
            raise LookupError(
                'backend parameters for "{0}" not found'.format(backend))

    ###############################################################
    # methods to compile quantum programs into qobj
    ###############################################################

    def compile(self, name_of_circuits, backend="local_qasm_simulator",
                config=None, silent=True, basis_gates=None, coupling_map=None,
                initial_layout=None, shots=1024, max_credits=10, seed=None,
                qobjid=None, hpc=None):
        """Compile the circuits into the exectution list.

        This builds the internal "to execute" list which is list of quantum
        circuits to run on different backends.

        Args:
            name_of_circuits (list[str]): circuit names to be compiled.
            backend (str): a string representing the backend to compile to
            config (dict): a dictionary of configurations parameters for the
                compiler
            silent (bool): is an option to print out the compiling information
                or not
            basis_gates (str): a comma seperated string and are the base gates,
                               which by default are: u1,u2,u3,cx,id
            coupling_map (dict): A directed graph of coupling::

                {
                 control(int):
                     [
                         target1(int),
                         target2(int),
                         , ...
                     ],
                     ...
                }

                eg. {0: [2], 1: [2], 3: [2]}

            initial_layout (dict): A mapping of qubit to qubit::

                                  {
                                    ("q", strart(int)): ("q", final(int)),
                                    ...
                                  }
                                  eg.
                                  {
                                    ("q", 0): ("q", 0),
                                    ("q", 1): ("q", 1),
                                    ("q", 2): ("q", 2),
                                    ("q", 3): ("q", 3)
                                  }

            shots (int): the number of shots
            max_credits (int): the max credits to use 3, or 5
            seed (int): the intial seed the simulatros use
            hpc (dict): This will setup some parameter for
                        ibmqx_hpc_qasm_simulator, using a JSON-like format like:
                        {
                            'multi_shot_optimization': Boolean,
                            'omp_num_threads': Numeric
                        }
                        This paramter MUST be used only with
                        ibmqx_hpc_qasm_simulator, otherwise the SDK will warn
                        the user via logging, and set the value to None.

        Returns:
            the job id and populates the qobj::

            qobj =
                {
                    id: --job id (string),
                    config: -- dictionary of config settings (dict)--,
                        {
                        "max_credits" (online only): -- credits (int) --,
                        "shots": -- number of shots (int) --.
                        "backend": -- backend name (str) --
                        }
                    circuits:
                        [
                            {
                            "name": --circuit name (string)--,
                            "compiled_circuit": --compiled quantum circuit (JSON format)--,
                            "compiled_circuit_qasm": --compiled quantum circuit (QASM format)--,
                            "config": --dictionary of additional config settings (dict)--,
                                {
                                "coupling_map": --adjacency list (dict)--,
                                "basis_gates": --comma separated gate names (string)--,
                                "layout": --layout computed by mapper (dict)--,
                                "seed": (simulator only)--initial seed for the simulator (int)--,
                                }
                            },
                            ...
                        ]
                    }

        """
        # TODO: Jay: currently basis_gates, coupling_map, initial_layout,
        # shots, max_credits and seed are extra inputs but I would like
        # them to go into the config.

        qobj = {}
        if not qobjid:
            qobjid = "".join([random.choice(string.ascii_letters+string.digits)
                              for n in range(30)])
        qobj['id'] = qobjid
        qobj["config"] = {"max_credits": max_credits, 'backend': backend,
                          "shots": shots}

        # TODO This backend needs HPC parameters to be passed in order to work
        if backend == 'ibmqx_hpc_qasm_simulator':
            if hpc is None:
                print('ibmqx_hpc_qasm_simulator backend needs HPC '
                      'parameter. Setting defaults to hpc.multi_shot_optimization '
                      '= true and hpc.omp_num_threads = 16')
                hpc = {'multi_shot_optimization': True, 'omp_num_threads': 16}

            if not all (key in hpc for key in
                ('multi_shot_optimization','omp_num_threads')):
                raise QISKitError('Unknown HPC parameter format!')

            qobj['config']['hpc'] = hpc
        elif hpc is not None:
            print('HPC paramter is only available for '
                  'ibmqx_hpc_qasm_simulator. You are passing an HPC parameter '
                  'but you are not using ibmqx_hpc_qasm_simulator, so we will '
                  'ignore it.')
            hpc = None

        qobj["circuits"] = []

        if not name_of_circuits:
            raise ValueError('"name_of_circuits" must be specified')
        if isinstance(name_of_circuits, str):
            name_of_circuits = [name_of_circuits]
        for name in name_of_circuits:
            if name not in self.__quantum_program:
                raise QISKitError('circuit "{0}" not found in program'.format(name))
            if not basis_gates:
                basis_gates = "u1,u2,u3,cx,id"  # QE target basis
            # TODO: The circuit object going into this is to have .qasm() method (be careful)
            dag_circuit = self._unroller_code(self.__quantum_program[name],
                                              basis_gates=basis_gates)
            final_layout = None
            # if a coupling map is given compile to the map
            if coupling_map:
                if not silent:
                    print("pre-mapping properties: %s"
                          % dag_circuit.property_summary())
                # Insert swap gates
                coupling = self.mapper.Coupling(coupling_map)
                if not silent:
                    print("initial layout: %s" % initial_layout)
                dag_circuit, final_layout = self.mapper.swap_mapper(
                    dag_circuit, coupling, initial_layout, trials=20, verbose=False)
                if not silent:
                    print("final layout: %s" % final_layout)
                # Expand swaps
                dag_circuit = self._unroller_code(dag_circuit)
                # Change cx directions
                dag_circuit = mapper.direction_mapper(dag_circuit, coupling)
                # Simplify cx gates
                mapper.cx_cancellation(dag_circuit)
                # Simplify single qubit gates
                dag_circuit = mapper.optimize_1q_gates(dag_circuit)
                if not silent:
                    print("post-mapping properties: %s"
                          % dag_circuit.property_summary())

            # making the job to be added to qojj
            job = {}
            job["name"] = name
            # config parameters used by the runner
            if config is None:
                config = {}  # default to empty config dict
            job["config"] = config
            # TODO: Jay: make config options optional for different backends
            job["config"]["coupling_map"] = mapper.coupling_dict2list(coupling_map)
            # Map the layout to a format that can be json encoded
            list_layout = None
            if final_layout:
                list_layout = [[k, v] for k, v in final_layout.items()]
            job["config"]["layout"] = list_layout
            job["config"]["basis_gates"] = basis_gates
            if seed is None:
                job["config"]["seed"] = None
            else:
                job["config"]["seed"] = seed
            # the compuled circuit to be run saved as a dag
            job["compiled_circuit"] = self._dag2json(dag_circuit)
            job["compiled_circuit_qasm"] = dag_circuit.qasm(qeflag=True)
            # add job to the qobj
            qobj["circuits"].append(job)
        return qobj

    def get_execution_list(self, qobj, verbose=False):
        """Print the compiled circuits that are ready to run.

        Args:
            verbose (bool): controls how much is returned.
        """
        if not qobj:
            if verbose:
                print("no exectuions to run")
        execution_list_all = {}
        execution_list = []
        if verbose:
            print("id: %s" % qobj['id'])
            print("backend: %s" % qobj['config']['backend'])
            print("qobj config:")
            for key in qobj['config']:
                if key != 'backend':
                    print(' '+ key + ': ' + str(qobj['config'][key]))
        for circuit in qobj['circuits']:
            execution_list.append(circuit["name"])
            if verbose:
                print('  circuit name: ' + circuit["name"])
                print('  circuit config:')
                for key in circuit['config']:
                    print('   '+ key + ': ' + str(circuit['config'][key]))
        return execution_list

    def get_compiled_configuration(self, qobj, name):
        """Get the compiled layout for the named circuit and backend.

        Args:
            name (str):  the circuit name
            qobj (str): the name of the qobj

        Returns:
            the config of the circuit.
        """
        try:
            for index in range(len(qobj["circuits"])):
                if qobj["circuits"][index]['name'] == name:
                    return qobj["circuits"][index]["config"]
        except KeyError:
            raise QISKitError('No compiled configurations for circuit "{0}"'.format(name))

    def get_compiled_qasm(self, qobj, name):
        """Print the compiled cricuit in qasm format.

        Args:
            qobj (str): the name of the qobj
            name (str): name of the quantum circuit

        """
        try:
            for index in range(len(qobj["circuits"])):
                if qobj["circuits"][index]['name'] == name:
                    return qobj["circuits"][index]["compiled_circuit_qasm"]
        except KeyError:
            raise QISKitError('No compiled qasm for circuit "{0}"'.format(name))

    def _dag2json(self, dag_circuit):
        """Make a Json representation of the circuit.

        Takes a circuit dag and returns json circuit obj. This is an internal
        function.

        Args:
            dag_ciruit (dag object): a dag representation of the circuit

        Returns:
            the json version of the dag
        """
        # TODO: Jay: I think this needs to become a method like .qasm()
        # for the DAG.
        circuit_string = dag_circuit.qasm(qeflag=True)
        basis_gates = "u1,u2,u3,cx,id"  # QE target basis
        unroller = unroll.Unroller(qasm.Qasm(data=circuit_string).parse(),
                                   unroll.JsonBackend(basis_gates.split(",")))
        json_circuit = unroller.execute()
        return json_circuit

    def _unroller_code(self, dag_ciruit, basis_gates=None):
        """ Unroll the code.

        Circuit is the circuit to unroll using the DAG representation.
        This is an internal function.

        Args:
            dag_ciruit (dag object): a dag representation of the circuit
            basis_gates (str): a comma seperated string and are the base gates,
                               which by default are: u1,u2,u3,cx,id
        Return:
            dag_ciruit (dag object): a dag representation of the circuit
                                     unrolled to basis gates
        """
        if not basis_gates:
            basis_gates = "u1,u2,u3,cx,id"  # QE target basis
        unrolled_circuit = unroll.Unroller(qasm.Qasm(data=dag_ciruit.qasm()).parse(),
                                           unroll.DAGBackend(basis_gates.split(",")))
        dag_circuit_unrolled = unrolled_circuit.execute()
        return dag_circuit_unrolled

    ###############################################################
    # methods to run quantum programs (run )
    ###############################################################

    def run(self, qobj, wait=5, timeout=60, silent=True):
        """Run a program (a pre-compiled quantum program).

        All input for run comes from qobj

        Args:
            qobj(dict): the dictionary of the quantum object to run
            wait (int): wait time is how long to check if the job is completed
            timeout (int): is time until the execution stops
            silent (bool): is an option to print out the running information or
            not

        Returns:
            status done and populates the internal __quantum_program with the
            data

        """
        backend = qobj['config']['backend']
        if not silent:
            print("running on backend: %s" % (backend))
        if backend in self.__ONLINE_BACKENDS:
            max_credits = qobj["config"]["max_credits"]
            shots = qobj["config"]["shots"]
            seed = qobj["circuits"][0]["config"]["seed"]
            jobs = []
            for job in qobj["circuits"]:
                jobs.append({'qasm': job["compiled_circuit_qasm"]})

            hpc = None
            if (qobj['config']['backend'] == 'ibmqx_hpc_qasm_simulator' and
                    'hpc' in qobj['config']):
                try:
                    # Use CamelCase when passing the hpc parameters to the API.
                    hpc = {
                        'multiShotOptimization':
                            qobj['config']['hpc']['multi_shot_optimization'],
                        'ompNumThreads':
                            qobj['config']['hpc']['omp_num_threads']
                    }
                except:
                    hpc = None

            output = self.__api.run_job(jobs, backend, shots=shots,
                                        max_credits=max_credits, seed=seed,
                                        hpc=hpc)
            if 'error' in output:
                raise ResultError(output['error'])
            qobj_result = self._wait_for_job(output['id'], wait=wait,
                                             timeout=timeout, silent=silent)
        else:
            # making a list of jobs just for local backends. Name is droped
            # but the list is made ordered
            jobs = []
            for job in qobj["circuits"]:
                jobs.append({"compiled_circuit": job["compiled_circuit"],
                             "config": {**job["config"], **qobj["config"]}})
            qobj_result = self._run_local_simulator(backend, jobs, silent)
        if qobj_result['status'] == 'COMPLETED':
            assert len(qobj["circuits"]) == len(qobj_result['result']), (
                'Internal error in QuantumProgram.run(), job_result')
        results = Result(qobj_result, qobj)
        return results

    def _wait_for_job(self, jobid, wait=5, timeout=60, silent=True):
        """Wait until all online ran jobs are 'COMPLETED'.

        Args:
            jobids:  is a list of id strings.
            wait (int):  is the time to wait between requests, in seconds
            timeout (int):  is how long we wait before failing, in seconds
            silent (bool): is an option to print out the running information or
            not

        Returns:
             Dictionary of form::

                 job_result_return =
                     [
                        {
                         "data": DATA,
                         "status": DATA,
                         },
                         ...
                     ]
        """
        timer = 0
        timeout_over = False
        job_result = self.__api.get_job(jobid)
        if 'status' not in job_result:
            from pprint import pformat
            raise QISKitError("get_job didn't return status: %s" % (pformat(job)))
        while job_result['status'] == 'RUNNING':
            if timer >= timeout:
                return {"status": "ERROR", "result": ["Time Out"]}
            time.sleep(wait)
            timer += wait
            if not silent:
                print("status = %s (%d seconds)" % (job_result['status'], timer))
            job_result = self.__api.get_job(jobid)

            if 'status' not in job_result:
                from pprint import pformat
                raise QISKitError("get_job didn't return status: %s" % (pformat(job_result)))
            if job_result['status'] == 'ERROR_CREATING_JOB' or job_result['status'] == 'ERROR_RUNNING_JOB':
                return {"status": "ERROR", "result": [job_result['status']]}

        # Get the results
        job_result_return = []
        for index in range(len(job_result["qasms"])):
            job_result_return.append({"data": job_result["qasms"][index]["data"],
                                      "status": job_result["qasms"][index]["status"]})
        return {'status': job_result['status'], 'result': job_result_return}

    def _run_local_simulator(self, backend, jobs, silent=True):
        """Run a program of compiled quantum circuits on the local machine.

        Args:
          backend (str): the name of the local simulator to run
          jobs: list of dicts {"compiled_circuit": simulator input data,
                "config": integer num shots}

        Returns:
          Dictionary of form,
          job_results =
            [
                {
                "data": DATA,
                "status": DATA,
                },
                ...
            ]
        """
        job_results = []
        for job in jobs:
            local_simulator = simulators.LocalSimulator(backend, job)
            local_simulator.run(silent=silent)
            this_result = local_simulator.result()
            job_results.append(this_result)
        return {'status': 'COMPLETED', 'result': job_results}

    def execute(self, name_of_circuits, backend="local_qasm_simulator",
                config=None, wait=5, timeout=60, silent=True, basis_gates=None,
                coupling_map=None, initial_layout=None, shots=1024,
                max_credits=10, seed=None, hpc=None):

        """Execute, compile, and run an array of quantum circuits).

        This builds the internal "to execute" list which is list of quantum
        circuits to run on different backends.

        Args:
            name_of_circuits (list[str]): circuit names to be compiled.
            backend (str): a string representing the backend to compile to
            config (dict): a dictionary of configurations parameters for the
                compiler
            wait (int): wait time is how long to check if the job is completed
            timeout (int): is time until the execution stops
            silent (bool): is an option to print out the compiling information
            or not
            basis_gates (str): a comma seperated string and are the base gates,
                               which by default are: u1,u2,u3,cx,id
            coupling_map (dict): A directed graph of coupling::

                                {
                                control(int):
                                    [
                                        target1(int),
                                        target2(int),
                                        , ...
                                    ],
                                    ...
                                }
                                eg. {0: [2], 1: [2], 3: [2]}
            initial_layout (dict): A mapping of qubit to qubit
                                  {
                                  ("q", strart(int)): ("q", final(int)),
                                  ...
                                  }
                                  eg.
                                  {
                                  ("q", 0): ("q", 0),
                                  ("q", 1): ("q", 1),
                                  ("q", 2): ("q", 2),
                                  ("q", 3): ("q", 3)
                                  }
            shots (int): the number of shots
            max_credits (int): the max credits to use 3, or 5
            seed (int): the intial seed the simulatros use
            hpc (dict): This will setup some parameter for
                        ibmqx_hpc_qasm_simulator, using a JSON-like format like:
                        {
                            'multi_shot_optimization': Boolean,
                            'omp_num_threads': Numeric
                        }
                        This paramter MUST be used only with
                        ibmqx_hpc_qasm_simulator, otherwise the SDK will warn
                        the user via logging, and set the value to None.

        Returns:
            status done and populates the internal __quantum_program with the
            data
        """
        # TODO: Jay: currently basis_gates, coupling_map, intial_layout, shots,
        # max_credits, and seed are extra inputs but I would like them to go
        # into the config
        qobj = self.compile(name_of_circuits, backend=backend, config=config,
                            silent=silent, basis_gates=basis_gates,
                            coupling_map=coupling_map,
                            initial_layout=initial_layout, shots=shots,
                            max_credits=max_credits, seed=seed,
                            hpc=hpc)
        result = self.run(qobj, wait=wait, timeout=timeout, silent=silent)
        return result


class Result(object):
    """ Result Class.

    Class internal properties.

    Methods to process the quantum program after it has been run

    Internal::

        qobj =  { -- the quantum object that was complied --}
        result =
            [
                {
                "data":
                    {  #### DATA CAN BE A DIFFERENT DICTIONARY FOR EACH BACKEND ####
                    "counts": {’00000’: XXXX, ’00001’: XXXXX},
                    "time"  : xx.xxxxxxxx
                    },
                "status": --status (string)--
                },
                ...
            ]
    """

    def __init__(self, qobj_result, qobj):
        self.__qobj = qobj
        self.__result = qobj_result

    def __str__(self):
        """Get the status of the run.

        Returns:
            the status of the results.
        """
        return self.__result['status']

    def __iadd__(self, other):
        """Append a Result object to current Result object.

        Arg:
            other (Result): a Result object to append.
        Returns:
            The current object with appended results.
        """
        if self.__qobj['config'] == other.__qobj['config']:
            if isinstance(self.__qobj['id'], str):
                self.__qobj['id'] = [self.__qobj['id']]
            self.__qobj['id'].append(other.__qobj['id'])
            self.__qobj['circuits'] += other.__qobj['circuits']
            self.__result['result'] += other.__result['result']
            return self
        else:
            raise QISKitError('Result objects have different configs and cannot be combined.')

    def __add__(self, other):
        """Combine Result objects.

        Note that the qobj id of the returned result will be the same as the
        first result.

        Arg:
            other (Result): a Result object to combine.
        Returns:
            A new Result object consisting of combined objects.
        """
        ret = copy.deepcopy(self)
        ret += other
        return ret

    def get_error(self):
        if self.__result['status'] == 'ERROR':
            return self.__result['result'][0]
        else:
            return None

    def get_ran_qasm(self, name):
        """Get the ran qasm for the named circuit and backend.

        Args:
            name (str): the name of the quantum circuit.

        Returns:
            A text version of the qasm file that has been run.
        """
        try:
            qobj = self.__qobj
            for index in range(len(qobj["circuits"])):
                if qobj["circuits"][index]['name'] == name:
                    return qobj["circuits"][index]["compiled_circuit_qasm"]
        except KeyError:
            raise QISKitError('No  qasm for circuit "{0}"'.format(name))

    def get_data(self, name):
        """Get the data of cicuit name.

        The data format will depend on the backend. For a real device it
        will be for the form::

            "counts": {’00000’: XXXX, ’00001’: XXXX},
            "time"  : xx.xxxxxxxx

        for the qasm simulators of 1 shot::

            'quantum_state': array([ XXX,  ..., XXX]),
            'classical_state': 0

        for the qasm simulators of n shots::

            'counts': {'0000': XXXX, '1001': XXXX}

        for the unitary simulators::

            'unitary': np.array([[ XX + XXj
                                   ...
                                   XX + XX]
                                 ...
                                 [ XX + XXj
                                   ...
                                   XX + XXj]]

        Args:
            name (str): the name of the quantum circuit.

        Returns:
            A dictionary of data for the different backends.
        """
        try:
            qobj = self.__qobj
            for index in range(len(qobj["circuits"])):
                if qobj["circuits"][index]['name'] == name:
                    return self.__result['result'][index]["data"]
        except KeyError:
            raise QISKitError('No data for circuit "{0}"'.format(name))

    def get_counts(self, name):
        """Get the histogram data of cicuit name.

        The data from the a qasm circuit is dictionary of the format
        {’00000’: XXXX, ’00001’: XXXXX}.

        Args:
            name (str): the name of the quantum circuit.
            backend (str): the name of the backend the data was run on.

        Returns:
            A dictionary of counts {’00000’: XXXX, ’00001’: XXXXX}.
        """
        try:
            return self.get_data(name)['counts']
        except KeyError:
            raise QISKitError('No counts for circuit "{0}"'.format(name))

    def average_data(self, name, observable):
        """Compute the mean value of an diagonal observable.

        Takes in an observable in dictionary format and then
        calculates the sum_i value(i) P(i) where value(i) is the value of
        the observable for state i.

        Args:
            name (str): the name of the quantum circuit
            obsevable (dict): The observable to be averaged over. As an example
            ZZ on qubits equals {"00": 1, "11": 1, "01": -1, "10": -1}

        Returns:
            a double for the average of the observable
        """
        counts = self.get_counts(name)
        temp = 0
        tot = sum(counts.values())
        for key in counts:
            if key in observable:
                temp += counts[key] * observable[key] / tot
        return temp


class ResultError(QISKitError):
    """Exceptions raised due to errors in result output.

    It may be better for the QISKit API to raise this exception.

    Args:
        error (dict): This is the error record as it comes back from
            the API. The format is like::

                error = {'status': 403,
                         'message': 'Your credits are not enough.',
                         'code': 'MAX_CREDITS_EXCEEDED'}
    """
    def __init__(self, error):
        self.status = error['status']
        self.message = error['message']
        self.code = error['code']
